'''
Usage: DEBKS anno -c circ_pos -g gene_anno -m genome [-o output]

Options:
    -h --help                   Show help message.
    -v --version                Show version.
    -c circ_pos                 File of circRNA position in tab format (the first three columns are chr, start, end).
    -g gene_anno                Tabix indexed gtf File of gene annotation.
    -m genome                   Fasta file of genome.
    -o output                   Output prefix [default: anno].
'''

import pysam,pandas as pd, numpy as np,time,os,sys,traceback

def fileCheck(file):
    for i in file:
        if not os.path.exists(i):
            sys.exit('ERROR: %s is not exist!!!' % i)

def readGTFfile(fileName):
    if not os.path.exists(fileName):
        sys.exit('ERROR: %s is not exist!!!' % fileName)
    try:
        tabixfile = pysam.TabixFile(fileName)
        return(tabixfile)
    except:
        sys.exit('ERROR: make sure %s is sorted and tabix indexed!!!' % fileName)

def readCircFile(fileName):
    if not os.path.exists(fileName):
        sys.exit('ERROR: %s is not exist!!!' % fileName)
    fff=open(fileName)
    fl=fff.readline().strip().split('\t')
    fff.close()
    if len(fl) <3:
        sys.exit('ERROR: The number of columns in %s is not enough!!!' % fileName)
    isHeader=None
    for i in fl[1]:
        if i not in ['0','1','2','3','4','5','6','7','8','9']:
            isHeader=0
            break
    mat=pd.read_csv(fileName,sep='\t',header=isHeader)
    return(mat)

def readFaFile(fileName):
    if not os.path.exists(fileName):
        sys.exit('ERROR: %s is not exist!!!' % fileName)
    fff=open(fileName)
    fl=fff.readline()
    fff.close()
    if fl[0]!='>':
        sys.exit('ERROR:  %s need to be Fasta format!!!' % fileName)
    try:
        faFile=pysam.FastaFile(fileName)
        return(faFile)
    except:
        sys.exit('ERROR: make sure %s is fai indexed!!!' % fileName)

def anno(options):
    gtfFile=options['-g']
    faFile=options['-m']
    outAnno=options['-o']+'_detail.txt'
    outLen=options['-o']+'_len.txt'
    gtfFile=readGTFfile(gtfFile)
    faFile=readFaFile(faFile)
    circPos=readCircFile(options['-c']).iloc[:,0:3]
    circPos.columns=['chr','start','end']
    stime=time.time()
    print('Begin %s annotation for %d circRNAs.' % (options['-c'],circPos.shape[0]))
    annoPos=pd.DataFrame(circPos.apply(lambda x: pos2ano(x,gtfFile,faFile),axis=1).tolist(),columns=['type','strand','gene_id','gene_name','subtype','exon_start','exon_end','len'])
    annoPos_com=pd.concat([circPos,annoPos],axis=1)
    annoPos_com.to_csv(outAnno,index=False,sep='\t')
    annoPos['len']=annoPos['len'].map(str).tolist()
    lenDf0=annoPos['len'].map(lambda x: x.split(';')[0])
    lenDf=pd.concat([circPos,lenDf0],axis=1)
    lenDf.to_csv(outLen,index=False,sep='\t')
    etime=time.time()-stime
    print('Finished in %d seconds.' % etime)

##### function part ####
def getSeq(chr,start,end,faFile):
    try:
        return(faFile.fetch(chr,start,end))
    except:
        sys.exit('ERROR: make sure circ_pos and genome is match!!!' )

def getGene(chr,start,end,gtfFile):
    geneDict={}
    geneID_list=[]
    geneName_list=[]
    geneID_Dict={}
    geneStrand_list=[]
    if chr not in gtfFile.contigs:
        return(['intergenic'])
    for gtf in gtfFile.fetch(chr, start, end+1,parser=pysam.asGTF()):
        if gtf.feature=='gene':
            geneID_list.append(gtf.gene_id)
            geneName_list.append(gtf.gene_name)
            geneStrand_list.append(gtf.strand)
            geneID_Dict[gtf.gene_id]=[gtf.start,gtf.end]
        if gtf.feature=='exon':
            if geneDict.__contains__(gtf.transcript_id):
                geneDict[gtf.transcript_id].append([gtf.start,gtf.end,gtf.gene_id,gtf.gene_name,gtf.strand])
            else:
                geneDict[gtf.transcript_id]=[]
                geneDict[gtf.transcript_id].append([gtf.start,gtf.end,gtf.gene_id,gtf.gene_name,gtf.strand])
    if len(geneID_list)==0:
        return(['intergenic'])
    elif len(geneDict.keys())==0:
        return(['intronic',geneID_list,geneName_list,geneStrand_list])
    else:
        annoArr={}
        annoScore={'intergenic': 1, 'intronic': 3, 'inExon': 4, 'isExon': 7}
        for i in geneDict.keys():
            annoArr[i]=''
            tmpTranArr=geneDict[i]
            tmpExoArr_1=tmpTranArr[0]
            tmpGene=geneID_Dict[tmpExoArr_1[2]]
            if tmpExoArr_1[0]<start:
                annoArr[i]='inExon'
            elif tmpExoArr_1[0]==start:
                annoArr[i]='isExon'
            else:
                if tmpGene[0]>start:
                    annoArr[i]='intergenic'
                else:
                    annoArr[i]='intronic'
            tmpExoArr_L=tmpTranArr[-1]
            if tmpExoArr_L[1]<end:
                if tmpGene[1]<end:
                    annoArr[i]+='-intergenic'
                else:
                    annoArr[i]+='-intronic'
            elif tmpExoArr_L[1]==end:
                annoArr[i]+='-isExon'
            else:
                annoArr[i]+='-inExon'
        proTran=[]
        maxScore=0
        for i in annoArr.keys():
            tmpType=annoArr[i].split('-')
            tmpScore=annoScore[tmpType[0]]+annoScore[tmpType[1]]
            if tmpScore>maxScore:
                maxScore=tmpScore
                proTran=[i]
            elif tmpScore==maxScore:
                proTran.append(i)
            else:
                continue
        detailStart_arr=[]
        detailEnd_arr=[]
        detailLen_arr=[]
        detailGeneID_arr=[]
        detailGeneName_arr=[]
        detailAnno_arr=[]
        detailStrand_arr=[]
        for i in proTran:
            tmpS=[]
            tmpE=[]
            tmpTranArr=geneDict[i]
            lenExon=len(tmpTranArr)
            if lenExon==1:
                tmpS.append(start)
                tmpE.append(end)
            for j in range(lenExon):
                if j==0:
                    tmpS.append(start)
                    tmpE.append(tmpTranArr[j][1])
                elif j==lenExon-1:
                    tmpS.append(tmpTranArr[j][0])
                    tmpE.append(end)
                else:
                    tmpS.append(tmpTranArr[j][0])
                    tmpE.append(tmpTranArr[j][1])
            detailStart_arr.append(','.join([str(i+1) for i in tmpS]))
            detailEnd_arr.append(','.join([str(i) for i in tmpE]))
            detailLen_arr.append(sum(np.array(tmpE)-np.array(tmpS)))
            detailGeneID_arr.append(tmpTranArr[0][2])
            detailGeneName_arr.append(tmpTranArr[0][3])
            detailAnno_arr.append(annoArr[i])
            detailStrand_arr.append(tmpTranArr[0][4])
        return(['gene',
                list(set(zip(detailAnno_arr,detailGeneID_arr,detailGeneName_arr, detailStrand_arr,detailStart_arr,detailEnd_arr,detailLen_arr)))])
                
def getGenePair(geneID,geneName):
    geneZip=list(set(zip(geneID,geneName)))
    geneID=[]
    geneName=[]
    for i in geneZip:
        geneID.append(i[0])
        geneName.append(i[1])
    return(geneID,geneName)

def getGeneStrandPair(strand,geneID,geneName):
    geneZip=list(set(zip(strand,geneID,geneName)))
    strand=[]
    geneID=[]
    geneName=[]
    for i in geneZip:
        strand.append(i[0])
        geneID.append(i[1])
        geneName.append(i[2])
    return(strand,geneID,geneName)

def splitGenePair(x):
    geneID=[]
    geneName=[]
    for i in x:
        geneID.append(i[1])
        geneName.append(i[2])
    return(geneID,geneName)

def splitGeneAll(x,id):
    subType=[]
    geneID=[]
    geneName=[]
    start=[]
    end=[]
    circLen=[]
    strandList=[]
    for i in id:
        subType.append(x[i][0])
        geneID.append(x[i][1])
        geneName.append(x[i][2])
        strandList.append(x[i][3])
        start.append(x[i][4])
        end.append(x[i][5])
        circLen.append(str(x[i][6]))
    return(subType,geneID,geneName,strandList,start,end,circLen)

def subType2Type(x):
    topType=[]
    exonList=['isExon-inExon','inExon-isExon']
    mixList=['isExon-intronic','intronic-isExon','inExon-intronic','intronic-inExon',
             'isExon-intergenic','intergenic-isExon','inExon-intergenic','intergenic-inExon']
    for i in x:
        arr=list(set(i.split('-')))
        if len(arr)==1:
            if arr[0]=='intronic':
                topType.append('intronic')
            elif arr[0][2:]=='Exon':
                topType.append('exonic')
            else:
                topType.append('unknown')
        else:
            if i in exonList:
                topType.append('exon')
            elif i in mixList:
                topType.append('mix')
            else:
                topType.append('unknown')
    return(topType)

def adjGene(chr,start,end,gtfFile,strand):
    anno=getGene(chr,start,end,gtfFile)
    #type; strand; geneID; geneName; subType; start; end; len
    if anno[0]=='intergenic':
        return('intergenic',strand,'.','.','intergenic',start+1,end,end-start)
    elif anno[0]=='intronic':# type == intronic or intergenic; subType: antisense
        strand_list=anno[3]
        if strand in ['+','-']:
            keepID=[]
            for i in range(len(strand_list)):
                if strand_list[i]==strand:
                    keepID.append(i)
            if len(keepID)==0:
                geneID,geneName=getGenePair(anno[1],anno[2])
                return('intergenic',strand,';'.join(geneID),';'.join(geneName),'antisense',start+1,end,end-start)
            else:
                geneID=[anno[1][i] for i in keepID]
                geneName=[anno[2][i] for i in keepID]
                geneID,geneName=getGenePair(geneID,geneName)
                if (len(set(geneID))==1):
                    return('intronic',strand,geneID[0],geneName[0],'intronic',start+1,end,end-start)
                return('intronic',strand,';'.join(geneID),';'.join(geneName),'intronic',start+1,end,end-start)
        else:
            strand_list,geneID,geneName=getGeneStrandPair(strand_list,anno[1],anno[2])
            if (len(set(geneID))==1):
                return('intronic',strand_list[0],geneID[0],geneName[0],'intronic',start+1,end,end-start)
            return('intronic',';'.join(strand_list),';'.join(geneID),';'.join(geneName),'intronic',start+1,end,end-start)
    else:
        if strand in ['+','-']:
            keepID=[]
            for i in range(len(anno[1])):
                if anno[1][i][3]==strand:
                    keepID.append(i)
            if len(keepID)==0:
                geneID,geneName=splitGenePair(anno[1])
                if(len(set(geneID))==1):
                    return('intergenic',strand,geneID[0],geneName[0],'antisense',start+1,end,end-start)
                return('intergenic',strand,';'.join(geneID),';'.join(geneName),'antisense',start+1,end,end-start)
            else:
                subType,geneID,geneName,strand_list,start,end,circLen=splitGeneAll(anno[1],keepID)
                topType=subType2Type(subType)
                if(len(set(geneID))==1):
                    return(topType[0],strand,geneID[0],geneName[0],subType[0],';'.join(start),';'.join(end),';'.join(circLen))
                return(';'.join(topType),strand,';'.join(geneID),';'.join(geneName),';'.join(subType),';'.join(start),';'.join(end),';'.join(circLen))
        else:
            keepID=range(len(anno[1]))
            subType,geneID,geneName,strand_list,start,end,circLen=splitGeneAll(anno[1],keepID)
            topType=subType2Type(subType)
            if(len(set(geneID))==1):
                return(topType[0],strand_list[0],geneID[0],geneName[0],subType[0],';'.join(start),';'.join(end),';'.join(circLen))
            return(';'.join(topType),';'.join(strand_list),';'.join(geneID),';'.join(geneName),';'.join(subType),';'.join(start),';'.join(end),';'.join(circLen))

def pos2ano(pos,gtfFile,faFile):
    chr=pos[0]
    start=pos[1]-1
    end=pos[2]-1
    sSeq=getSeq(chr,start-2,start,faFile)
    eSeq=getSeq(chr,end+1,end+3,faFile)
    motifSeq=sSeq+eSeq
    strand='.'
    if motifSeq=='ACCT' or motifSeq=='GCCT' or motifSeq=='ATGT':
        strand='-'
    elif motifSeq=='AGGT' or motifSeq=='AGGC' or motifSeq=='ACAT':
        strand='+'
    return(list(adjGene(chr,start,end+1,gtfFile,strand)))



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# import libraries
import subprocess
import re,os,warnings,math,scipy,itertools,logging,time,datetime,pysam,traceback
from scipy import stats
from multiprocessing import Pool
from scipy.optimize import fmin_cobyla
from scipy.optimize import fmin_l_bfgs_b
from math import log
import pandas as pd
import numpy as np
from numpy import *
warnings.filterwarnings('ignore');
np.seterr(divide='ignore',invalid='ignore')
rho=0.9
np.random.seed(1231);
########## functions ############
#binomial MLE optimization functions
def logit(x):
    if x<0.01:
        x=0.01;
    if x>0.99:
        x=0.99;
    return(log(x/(1-x)));
    
def veclogit(x):
    res=[];
    for i in x:
        res.append(logit(i));
    return(res);

#START: log Marginal distribution for the paired analysis
#this is the full log likelihood function within the Laplace approximiation of the marginal distribution of paired replicates
#for one exon only
def myfunc_marginal_full_paired(x, *args):
    beta1=x[0];beta2=x[1];rho=x[2];var1=x[3];var2=x[4];
    I1=args[0];S1=args[1];psi1=args[2];
    I2=args[3];S2=args[4];psi2=args[5];
    inclusion_length=args[6];skipping_length=args[7];
    sigma=matrix([[var1,rho*sqrt(var1)*sqrt(var2)],[rho*sqrt(var1)*sqrt(var2),var2]]);
    sigma_inv=sigma.getI();
    vec=matrix([[logit(psi1)-logit(beta1)],[logit(psi2)-logit(beta2)]]);
    new_psi1=inclusion_length*psi1/(inclusion_length*psi1+skipping_length*(1-psi1));
    new_psi2=inclusion_length*psi2/(inclusion_length*psi2+skipping_length*(1-psi2));
    ##print('vec');#print(vec);
    temp1=0;temp2=0;temp3=0;temp4=0;
    temp1+=-0.5*log(abs(linalg.det(sigma)));
    temp2+=-0.5*dot(dot(vec.getT(),sigma_inv),vec);
    temp3+=I1*log(new_psi1)+S1*log(1-new_psi1)+log(psi1)+log(1-psi1);
    temp4+=I2*log(new_psi2)+S2*log(1-new_psi2)+log(psi2)+log(1-psi2);
    return(temp1+temp2+temp3+temp4);

#return the det of sigma1 matrix for one exon
def myfunc_marginal_sigma1_paired(x, *args):
    beta1=x[0];beta2=x[1];rho=x[2];var1=x[3];var2=x[4];
    I1=args[0];S1=args[1];psi1=args[2];
    I2=args[3];S2=args[4];psi2=args[5];
    inclusion_length=args[6];skipping_length=args[7];
    sigma=matrix([[var1,rho*sqrt(var1)*sqrt(var2)],[rho*sqrt(var1)*sqrt(var2),var2]]);
    sigma_inv=sigma.getI();
    vec=matrix([[logit(psi1)-logit(beta1)],[logit(psi2)-logit(beta2)]]);
    vec_der=matrix([[1/psi1/(1-psi1)],[1/psi2/(1-psi2)]]);
    vec_2der=matrix([[(-1+2*psi1)/pow(psi1,2)/pow(1-psi1,2)],[(-1+2*psi2)/pow(psi2,2)/pow(1-psi2,2)]]);
    new_psi1=inclusion_length*psi1/(inclusion_length*psi1+skipping_length*(1-psi1));
    new_psi2=inclusion_length*psi2/(inclusion_length*psi2+skipping_length*(1-psi2));
    a_apostrophe=sigma_inv[0,0]*pow(vec_der[0,0],2)+(sigma_inv[0,0]*vec[0,0]+sigma_inv[0,1]*vec[1,0])*vec_2der[0,0];
    a_apostrophe+=-1*pow(psi1,-2)+pow(1-psi1,-2);
    a_apostrophe+=-1*I1*skipping_length*((2*inclusion_length+skipping_length)*psi1+skipping_length*(1-psi1))/pow(psi1,2)/pow(inclusion_length*psi1+skipping_length*(1-psi1),2);
    a_apostrophe+=-1*S1*inclusion_length*((inclusion_length+2*skipping_length)*(1-psi1)+inclusion_length*psi1)/pow(1-psi1,2)/pow(inclusion_length*psi1+skipping_length*(1-psi1),2);
    b_apostrophe=sigma_inv[0,1]*vec_der[0,0]*vec_der[1,0];
    d_apostrophe=sigma_inv[1,1]*pow(vec_der[1,0],2)+(sigma_inv[1,1]*vec[1,0]+sigma_inv[0,1]*vec[0,0])*vec_2der[1,0];
    d_apostrophe+=-1*pow(psi2,-2)+pow(1-psi2,-2);
    d_apostrophe+=-1*I2*skipping_length*((2*inclusion_length+skipping_length)*psi2+skipping_length*(1-psi2))/pow(psi2,2)/pow(inclusion_length*psi2+skipping_length*(1-psi2),2);
    d_apostrophe+=-1*S2*inclusion_length*((inclusion_length+2*skipping_length)*(1-psi2)+inclusion_length*psi2)/pow(1-psi2,2)/pow(inclusion_length*psi2+skipping_length*(1-psi2),2);
    sigma_1=matrix([[a_apostrophe,b_apostrophe],[b_apostrophe,d_apostrophe]]);
    return(-1*sigma_1);

def myfunc_marginal_paired_rho0(x, *args):
    beta1=x[0];beta2=x[1];rho=0;
    I1=args[0];S1=args[1];psi1=args[2];
    I2=args[3];S2=args[4];psi2=args[5];
    inclusion_length=args[6];skipping_length=args[7];
    var1=args[8];var2=args[9];
    sum=0;temp1=matrix([[1]]);temp2=0;
    sum_temp1=0;sum_temp2=0;
    for i in range(len(psi1)):
        temp1=myfunc_marginal_sigma1_paired([beta1,beta2,rho,var1,var2],I1[i],S1[i],psi1[i],I2[i],S2[i],psi2[i],inclusion_length,skipping_length);
        temp2=myfunc_marginal_full_paired([beta1,beta2,rho,var1,var2],I1[i],S1[i],psi1[i],I2[i],S2[i],psi2[i],inclusion_length,skipping_length);
        sum+=(-0.5*log(abs(linalg.det(temp1))+1)+temp2);
        sum_temp1+=temp1;sum_temp2+=temp2;
    return(-1*sum);

def myfunc_marginal_der_paired_rho0(x, *args):
    beta1=x[0];beta2=x[1];rho=0;
    I1=args[0];S1=args[1];psi1=args[2];
    I2=args[3];S2=args[4];psi2=args[5];
    inclusion_length=args[6];skipping_length=args[7];
    var1=args[8];var2=args[9];
    sum=array([0.0,0.0]);
    sum_temp1=array([0.0,0.0]);sum_temp2=array([0.0,0.0]);
    temp1=0;temp2=0;
    for i in range(len(psi1)):
        temp1=myfunc_marginal_sigma1_der_paired([beta1,beta2,rho,var1,var2],I1[i],S1[i],psi1[i],I2[i],S2[i],psi2[i],inclusion_length,skipping_length,1);
        temp2=myfunc_marginal_full_der_paired([beta1,beta2,rho,var1,var2],I1[i],S1[i],psi1[i],I2[i],S2[i],psi2[i],inclusion_length,skipping_length);
        sum_temp1+=temp1[0:2];sum_temp2+=temp2[0:2];
        sum=sum+temp1[0:2]+temp2[0:2];
        sigma1=myfunc_marginal_sigma1_paired([beta1,beta2,rho,var1,var2],I1[i],S1[i],psi1[i],I2[i],S2[i],psi2[i],inclusion_length,skipping_length);
        #if len(args)>10:
        #    ofile=open('test_output.txt','a');ofile.write(str(temp1[0])+'\t'+str(temp1[1])+'\t'+str(psi1[i])+'\t'+str(psi2[i])+'\t'+str(linalg.det(sigma1))+'\t'+str(-0.5/(linalg.det(sigma1)+1))+'\t'+str(list(sigma1))+'\n');
    return(-1*array(sum));    
    
def myfunc_marginal_paired_rho(x, *args):
    beta1=x[0];beta2=x[1];rho=x[2];
    I1=args[0];S1=args[1];psi1=args[2];
    I2=args[3];S2=args[4];psi2=args[5];
    inclusion_length=args[6];skipping_length=args[7];
    var1=args[8];var2=args[9];
    sum=0;temp1=matrix([[1]]);temp2=0;
    sum_temp1=0;sum_temp2=0;
    for i in range(len(psi1)):
        temp1=myfunc_marginal_sigma1_paired([beta1,beta2,rho,var1,var2],I1[i],S1[i],psi1[i],I2[i],S2[i],psi2[i],inclusion_length,skipping_length);
        temp2=myfunc_marginal_full_paired([beta1,beta2,rho,var1,var2],I1[i],S1[i],psi1[i],I2[i],S2[i],psi2[i],inclusion_length,skipping_length);
        sum+=(-0.5*log(abs(linalg.det(temp1))+1)+temp2);
    return(-1*sum);

def myfunc_marginal_der_paired_rho(x, *args):
    beta1=x[0];beta2=x[1];rho=x[2];
    I1=args[0];S1=args[1];psi1=args[2];
    I2=args[3];S2=args[4];psi2=args[5];
    inclusion_length=args[6];skipping_length=args[7];
    var1=args[8];var2=args[9];
    sum=array([0.0,0.0,0.0]);
    sum_temp1=array([0.0,0.0,0.0,0.0,0.0]);sum_temp2=array([0.0,0.0,0.0,0.0,0.0]);
    temp1=0;temp2=0;
    for i in range(len(psi1)):
        temp1=myfunc_marginal_sigma1_der_paired([beta1,beta2,rho,var1,var2],I1[i],S1[i],psi1[i],I2[i],S2[i],psi2[i],inclusion_length,skipping_length);
        temp2=myfunc_marginal_full_der_paired([beta1,beta2,rho,var1,var2],I1[i],S1[i],psi1[i],I2[i],S2[i],psi2[i],inclusion_length,skipping_length);
        sum=sum+(temp1[0:3]+temp2[0:3]);
    return(-1*array(sum));
    
def myfunc_marginal_paired_fixvar(x, *args):
    beta1=x[0];beta2=x[1];
    #beta1=x[0];beta2=args[11];
    #beta2=x[0];beta1=args[11];
    rho=args[8];var1=args[9];var2=args[10];
    I1=args[0];S1=args[1];psi1=args[2];
    I2=args[3];S2=args[4];psi2=args[5];
    inclusion_length=args[6];skipping_length=args[7];
    sum=0;temp1=matrix([[1]]);temp2=0;
    sum_temp1=0;sum_temp2=0;
    ##print('fixvar_beta1');#print(beta1);#print('fixvar_beta2');#print(beta2);
    for i in range(len(psi1)):
        temp1=myfunc_marginal_sigma1_paired([beta1,beta2,rho,var1,var2],I1[i],S1[i],psi1[i],I2[i],S2[i],psi2[i],inclusion_length,skipping_length);
        temp2=myfunc_marginal_full_paired([beta1,beta2,rho,var1,var2],I1[i],S1[i],psi1[i],I2[i],S2[i],psi2[i],inclusion_length,skipping_length);
        sum+=(0.5*log(abs(linalg.det(temp1))+1)-temp2);
        sum_temp1+=temp1;sum_temp2+=temp2;
    ##print('sum_temp1');#print(sum_temp1);#print('sum_temp2');#print(sum_temp2);#print('sum');#print(sum);
    return(sum);

def myfunc_marginal_der_paired_fixvar(x, *args):
    beta1=x[0];beta2=x[1];
    #beta1=x[0];beta2=args[11];
    #beta2=x[0];beta1=args[11];
    rho=args[8];var1=args[9];var2=args[10];
    I1=args[0];S1=args[1];psi1=args[2];
    I2=args[3];S2=args[4];psi2=args[5];
    inclusion_length=args[6];skipping_length=args[7];
    sum=array([0.0,0.0]);
    #sum_temp1=array([0.0,0.0,0.0,0.0,0.0]);sum_temp2=array([0.0,0.0,0.0,0.0,0.0]);
    temp1=0;temp2=0;
    for i in range(len(psi1)):
        temp1=myfunc_marginal_sigma1_der_paired([beta1,beta2,rho,var1,var2],I1[i],S1[i],psi1[i],I2[i],S2[i],psi2[i],inclusion_length,skipping_length);
        temp2=myfunc_marginal_full_der_paired([beta1,beta2,rho,var1,var2],I1[i],S1[i],psi1[i],I2[i],S2[i],psi2[i],inclusion_length,skipping_length);
        #sum_temp1=sum_temp1-1*temp1;sum_temp2=sum_temp2-1*temp2;
        sum=sum+(-1*temp1[0:2]-1*temp2[0:2]);
    ##print('sum_temp1_der');#print(sum_temp1);#print('sum_temp2_der');#print(sum_temp2);
    return(array(sum));

def myfunc_marginal_paired(x, *args):
    beta1=x[0];beta2=x[1];rho=x[2];var1=x[3];var2=x[4];
    I1=args[0];S1=args[1];psi1=args[2];
    I2=args[3];S2=args[4];psi2=args[5];
    inclusion_length=args[6];skipping_length=args[7];
    sum=0;temp1=matrix([[1]]);temp2=0;
    for i in range(len(psi1)):
        temp1=myfunc_marginal_sigma1_paired([beta1,beta2,rho,var1,var2],I1[i],S1[i],psi1[i],I2[i],S2[i],psi2[i],inclusion_length,skipping_length);
        temp2=myfunc_marginal_full_paired([beta1,beta2,rho,var1,var2],I1[i],S1[i],psi1[i],I2[i],S2[i],psi2[i],inclusion_length,skipping_length);
        sum+=(-0.5*log(abs(linalg.det(temp1))+1)+temp2);
    return(-1*sum);
#END: log Marginal distribution for the paired analysis

#START: Derivative of the log Marginal distribution for the paired analysis
def myfunc_marginal_full_der_paired(x, *args):
    beta1=x[0];beta2=x[1];rho=x[2];var1=x[3];var2=x[4];
    I1=args[0];S1=args[1];psi1=args[2];
    I2=args[3];S2=args[4];psi2=args[5];
    inclusion_length=args[6];skipping_length=args[7];
    sigma=matrix([[var1,rho*sqrt(var1)*sqrt(var2)],[rho*sqrt(var1)*sqrt(var2),var2]]);
    sigma_inv=sigma.getI();
    vec=matrix([[logit(psi1)-logit(beta1)],[logit(psi2)-logit(beta2)]]);
    temp=-1*dot(transpose(vec),dot(sigma_inv,matrix([[-1/beta1/(1-beta1),0],[0,-1/beta2/(1-beta2)]])));
    temp1=temp[0,0];temp2=temp[0,1];
    diag=(-2*rho-1+pow(rho,2))/sqrt(var1)/sqrt(var2);
    temp3=-0.5*dot(transpose(vec),dot(matrix([[2*rho/var1,diag],[diag,2*rho/var2]]),vec))/pow(1-rho*rho,2)-0.5/linalg.det(sigma)*(-1*var1*var2);
    #temp4=(-1/pow(var1,2)*pow(vec[0,0],2)+rho/pow(var1,1.5)/pow(var2,0.5)*vec[0,0]*vec[1,0])/(1-rho*rho)+(1-rho*rho)*var2/linalg.det(sigma);
    #temp5=(-1/pow(var2,2)*pow(vec[1,0],2)+rho/pow(var2,1.5)/pow(var1,0.5)*vec[0,0]*vec[1,0])/(1-rho*rho)+(1-rho*rho)*var1/linalg.det(sigma);
    return(array([float(temp1),float(temp2),float(temp3)]));

def myfunc_marginal_sigma1_der_paired(x, *args):
    beta1=x[0];beta2=x[1];rho=x[2];var1=x[3];var2=x[4];
    I1=args[0];S1=args[1];psi1=args[2];
    I2=args[3];S2=args[4];psi2=args[5];
    inclusion_length=args[6];skipping_length=args[7];
    sigma=matrix([[var1,rho*sqrt(var1)*sqrt(var2)],[rho*sqrt(var1)*sqrt(var2),var2]]);
    sigma_inv=sigma.getI();
    ##print('sigma_inv');#print(sigma_inv);
    vec=matrix([[logit(psi1)-logit(beta1)],[logit(psi2)-logit(beta2)]]);
    vec_der=matrix([[1/psi1/(1-psi1)],[1/psi2/(1-psi2)]]);
    vec_2der=matrix([[(-1+2*psi1)/pow(psi1,2)/pow(1-psi1,2)],[(-1+2*psi2)/pow(psi2,2)/pow(1-psi2,2)]]);
    sigma1=myfunc_marginal_sigma1_paired([beta1,beta2,rho,var1,var2],I1,S1,psi1,I2,S2,psi2,inclusion_length,skipping_length);
    temp1=sigma_inv[0,0]*(-1/beta1/(1-beta1))*vec_2der[0,0]*sigma1[1,1]*(-1)+sigma_inv[0,1]*(-1/beta1/(1-beta1))*vec_2der[1,0]*sigma1[0,0]*(-1);
    temp2=sigma_inv[1,1]*(-1/beta2/(1-beta2))*vec_2der[1,0]*sigma1[0,0]*(-1)+sigma_inv[0,1]*(-1/beta2/(1-beta2))*vec_2der[0,0]*sigma1[1,1]*(-1);
    #temp1=(-1/var1/beta1/(1-beta1)*vec_2der[0,0])*sigma1[1,1]/(1-rho*rho)*(-1)-(rho/sqrt(var1)/sqrt(var2)/beta1/(1-beta1)*vec_2der[1,0])*sigma1[0,0]/(1-rho*rho);
    #temp2=(-1/var2/beta2/(1-beta2)*vec_2der[1,0])*sigma1[0,0]/(1-rho*rho)*(-1)-(rho/sqrt(var1)/sqrt(var2)/beta2/(1-beta2)*vec_2der[0,0])*sigma1[1,1]/(1-rho*rho);
    
    temp3=(-1/sqrt(var1)/sqrt(var2)*vec[1,0]*vec_2der[0,0])*sigma1[1,1]/(1-rho*rho)*(-1);
    temp3+=(-1/sqrt(var1)/sqrt(var2)*vec[0,0]*vec_2der[1,0])*sigma1[0,0]/(1-rho*rho)*(-1);
    temp3+=-2*(-1/sqrt(var1)/sqrt(var2)*vec_der[0,0]*vec_der[1,0])*sigma1[0,1]/(1-rho*rho)*(-1);
    temp3+=4*rho/(1-rho*rho)*linalg.det(sigma1);
    ##print('test_der');#print(linalg.det(sigma1));#print(x);#print(-0.5/(linalg.det(sigma1))*array([float(temp1),float(temp2),float(temp3)]));
    #if len(args)>8:
    #    ofile=open('test_output_2.txt','a');ofile.write(str([float(temp1),float(temp2),float(temp3)])+'\n');
    if linalg.det(sigma1)>0:
        return(-0.5/(linalg.det(sigma1)+1)*array([float(temp1),float(temp2),float(temp3)]));
    else:
        return(-0.5/(linalg.det(sigma1)-1)*array([float(temp1),float(temp2),float(temp3)]));
        
def myfunc_marginal_der_paired(x, *args):
    beta1=x[0];beta2=x[1];rho=x[2];var1=x[3];var2=x[4];
    I1=args[0];S1=args[1];psi1=args[2];
    I2=args[3];S2=args[4];psi2=args[5];
    inclusion_length=args[6];skipping_length=args[7];
    sum=array([0.0,0.0,0.0,0.0,0.0]);
    sum_temp1=array([0.0,0.0,0.0,0.0,0.0]);sum_temp2=array([0.0,0.0,0.0,0.0,0.0]);
    temp1=0;temp2=0;
    for i in range(len(psi1)):
        temp1=myfunc_marginal_sigma1_der_paired([beta1,beta2,rho,var1,var2],I1[i],S1[i],psi1[i],I2[i],S2[i],psi2[i],inclusion_length,skipping_length);
        temp2=myfunc_marginal_full_der_paired([beta1,beta2,rho,var1,var2],I1[i],S1[i],psi1[i],I2[i],S2[i],psi2[i],inclusion_length,skipping_length);
        sum_temp1=sum_temp1-1*temp1;sum_temp2=sum_temp2-1*temp2;
        sum=sum+(-1*temp1-1*temp2);
    return(array(sum[0:5]));
#END: Derivative of the log Marginal distribution for the paired analysis
            

#START: log-likelihood function of each individual replicate
def myfunc_individual_paired(x, *args):
    psi1=x[0];psi2=x[1];
    I1=args[0];S1=args[1];beta1=args[2];var1=args[3];
    I2=args[4];S2=args[5];beta2=args[6];var2=args[7];
    inclusion_length=args[8];skipping_length=args[9];
    rho=args[10];
    temp=myfunc_marginal_full_paired([beta1,beta2,rho,var1,var2],I1,S1,psi1,I2,S2,psi2,inclusion_length,skipping_length);
    return(-1*temp);  

def myfunc_individual_der_paired(x,*args):
    psi1=x[0];psi2=x[1];
    I1=args[0];S1=args[1];beta1=args[2];var1=args[3];
    I2=args[4];S2=args[5];beta2=args[6];var2=args[7];
    inclusion_length=args[8];skipping_length=args[9];
    rho=args[10];
    sigma=matrix([[var1,rho*sqrt(var1)*sqrt(var2)],[rho*sqrt(var1)*sqrt(var2),var2]]);
    sigma_inv=sigma.getI();
    vec=matrix([[logit(psi1)-logit(beta1)],[logit(psi2)-logit(beta2)]]);
    vec_der=matrix([[1/psi1/(1-psi1)],[1/psi2/(1-psi2)]]);
    new_psi1=inclusion_length*psi1/(inclusion_length*psi1+skipping_length*(1-psi1));
    new_psi1_der=inclusion_length*skipping_length/pow(inclusion_length*psi1+skipping_length*(1-psi1),2);
    new_psi2=inclusion_length*psi2/(inclusion_length*psi2+skipping_length*(1-psi2));
    new_psi2_der=inclusion_length*skipping_length/pow(inclusion_length*psi2+skipping_length*(1-psi2),2);
    temp1=0;temp2=0;
    temp1+=-1*(sigma_inv[0,0]*vec[0,0]*vec_der[0,0]+sigma_inv[0,1]*vec_der[0,0]*vec[1,0]);
    temp1+=1*(I1/new_psi1*new_psi1_der-S1/(1-new_psi1)*new_psi1_der+1/psi1-1/(1-psi1));
    temp2+=-1*(sigma_inv[1,1]*vec[1,0]*vec_der[1,0]+sigma_inv[0,1]*vec_der[1,0]*vec[0,0]);
    temp2+=1*(I2/new_psi2*new_psi2_der-S2/(1-new_psi2)*new_psi2_der+1/psi2-1/(1-psi2));
    return(-1*array([temp1,temp2]));
#END: log-likelihood function of each individual replicate
    
def myfunc_multivar(x,*args):
    psi1=args[0];psi2=args[1];var1=args[2];var2=args[3];
    ##print('psi1');#print(psi1);#print('psi2');#print(psi2);
    sum1=0;sum2=0;
    for i in psi1:
        sum1+=pow(logit(i)-logit(x[0]),2);
    sum1=sum1/var1/2;
    for i in psi2:
        sum2+=pow(logit(i)-logit(x[1]),2);
    sum2=sum2/var2/2;
    return(sum1+sum2+0.1*0.5*(pow(rho,2))/(1-pow(rho,2))*(pow(stats.norm.ppf(x[0]),2)+pow(stats.norm.ppf(x[1]),2)-2*rho*stats.norm.ppf(x[0])*stats.norm.ppf(x[1])));

def myfunc_multivar_der(x,*args):
    psi1=args[0];psi2=args[1];var1=args[2];var2=args[3];
    sum1=0;sum2=0;
    for i in psi1:
        sum1+=-2*(logit(i)-logit(x[0]))/x[0]/(1-x[0]);
    sum1=sum1/var1/2;
    res1=sum1+0.1*0.5*(pow(rho,2))/(1-pow(rho,2))*(2*stats.norm.ppf(x[0])-2*rho*stats.norm.ppf(x[1]))/stats.norm.pdf(stats.norm.ppf(x[0]));
    ##print('1');#print(x[1]);#print(res1);#print(stats.norm.pdf(stats.norm.ppf(x[0])));
    for i in psi2:
        sum2+=-2*(logit(i)-logit(x[1]))/x[1]/(1-x[1]);
    sum2=sum2/var2/2;
    res2=sum2+0.1*0.5*(pow(rho,2))/(1-pow(rho,2))*(2*stats.norm.ppf(x[1])-2*rho*stats.norm.ppf(x[0]))/stats.norm.pdf(stats.norm.ppf(x[1]));
    return(np.array([res1,res2]));

def myfunc_1(x, *args):
    psi1=args[0];psi2=args[1];var1=args[2];var2=args[3];
    sum1=0;sum2=0;
    for i in psi1:
        sum1+=pow(logit(i)-logit(x+cutoff),2);
    sum1=sum1/var1/2;
    for i in psi2:
        sum2+=pow(logit(i)-logit(x),2);
    sum2=sum2/var2/2;
    return(sum1+sum2+0.1*0.5*(pow(rho,2))/(1-pow(rho,2))*(pow(stats.norm.ppf(x+cutoff),2)+pow(stats.norm.ppf(x),2)-2*rho*stats.norm.ppf(x+cutoff)*stats.norm.ppf(x)));

def myfunc_der_1(x, *args):
    psi1=args[0];psi2=args[1];var1=args[2];var2=args[3];
    sum1=0;sum2=0;
    for i in psi1:
        sum1+=-2*(logit(i)-logit(x+cutoff))/(x+cutoff)/(1-x-cutoff);
    sum1=sum1/var1/2;
    res1=sum1+0.1*0.5*(pow(rho,2))/(1-pow(rho,2))*(2*stats.norm.ppf(x+cutoff)-2*rho*stats.norm.ppf(x))/stats.norm.pdf(stats.norm.ppf(x+cutoff));
    for i in psi2:
        sum2+=-2*(logit(i)-logit(x))/x/(1-x);
    sum2=sum2/var2/2;
    res2=sum2+0.1*0.5*(pow(rho,2))/(1-pow(rho,2))*(2*stats.norm.ppf(x)-2*rho*stats.norm.ppf(x+cutoff))/stats.norm.pdf(stats.norm.ppf(x));
    return(res1+res2);

def myfunc_2(x, *args):
    psi1=args[0];psi2=args[1];var1=args[2];var2=args[3];
    sum1=0;sum2=0;
    for i in psi1:
        sum1+=pow(logit(i)-logit(x),2);
    sum1=sum1/var1/2;
    for i in psi2:
        sum2+=pow(logit(i)-logit(x+cutoff),2);
    sum2=sum2/var2/2;
    return(sum1+sum2+0.1*0.5*(pow(rho,2))/(1-pow(rho,2))*(pow(stats.norm.ppf(x+cutoff),2)+pow(stats.norm.ppf(x),2)-2*rho*stats.norm.ppf(x+cutoff)*stats.norm.ppf(x)));

def myfunc_der_2(x, *args):
    psi1=args[0];psi2=args[1];var1=args[2];var2=args[3];
    sum1=0;sum2=0;
    for i in psi1:
        sum1+=-2*(logit(i)-logit(x))/(x)/(1-x);
    sum1=sum1/var1/2;
    res1=sum1+0.1*0.5*(pow(rho,2))/(1-pow(rho,2))*(2*stats.norm.ppf(x+cutoff)-2*rho*stats.norm.ppf(x))/stats.norm.pdf(stats.norm.ppf(x+cutoff));
    for i in psi2:
        sum2+=-2*(logit(i)-logit(x+cutoff))/(x+cutoff)/(1-x-cutoff);
    sum2=sum2/var2/2;
    res2=sum2+0.1*0.5*(pow(rho,2))/(1-pow(rho,2))*(2*stats.norm.ppf(x)-2*rho*stats.norm.ppf(x+cutoff))/stats.norm.pdf(stats.norm.ppf(x));
    return(res1+res2);

def myfunc_marginal_2der(x, args):
    I=args[0];S=args[1];beta=args[2];var=args[3];
    inclusion_length=args[4];
    skipping_length=args[5]
    temp1=1/pow(x,2)/pow(1-x,2)*(((2*x-1)*(logit(beta)-logit(x))-1)/var-1);#temp1=(2*x-1)/pow(x,2)/pow(1-x,2)*((logit(beta)-logit(x)-1/(2*x-1))/var+1);
    temp2=I*skipping_length*((2*inclusion_length+skipping_length)*x+skipping_length*(1-x))/pow(x,2)/pow(inclusion_length*x+skipping_length*(1-x),2);
    temp3=S*inclusion_length*((inclusion_length+2*skipping_length)*(1-x)+inclusion_length*x)/pow(1-x,2)/pow(inclusion_length*x+skipping_length*(1-x),2);
    ##print('test');#print(beta);#print(x);#print(var);#print(temp1);#print(temp2);#print(temp3);
    return(temp1-temp2-temp3);
    
def myfunc_marginal(x, *args):
    beta=x;
    I=args[0];S=args[1];psi=args[2];var=args[3];
    inclusion_length=args[4];skipping_length=args[5];
    sum=0;temp1=0;temp2=0;
    for i in range(len(psi)):
        new_psi=inclusion_length*psi[i]/(inclusion_length*psi[i]+skipping_length*(1-psi[i]));
        f1=I[i]*log(new_psi)+S[i]*log(1-new_psi)-pow(logit(psi[i])-logit(beta),2)/2/var-log(psi[i])-log(1-psi[i]);
        temp2+=-1*f1;
        f1_2der=abs(myfunc_marginal_2der(psi[i],[I[i],S[i],beta,var,inclusion_length,skipping_length]));
        temp1+=0.5*log(abs(f1_2der));
        sum+=(0.5*log(abs(f1_2der)+0.00001)-f1);
    ##print('test_temp1');#print(temp1);#print(temp2);
    return(sum);

def myfunc_marginal_der(x, *args):
    beta=x;
    I=args[0];S=args[1];psi=args[2];var=args[3];
    inclusion_length=args[4];skipping_length=args[5];
    sum=0;
    for i in range(len(psi)):
        new_psi=inclusion_length*psi[i]/(inclusion_length*psi[i]+skipping_length*(1-psi[i]));
        f1_3der=1*(2*psi[i]-1)/pow(psi[i],2)/pow(1-psi[i],2)/beta/(1-beta)/var;
        f1_2der=myfunc_marginal_2der(psi[i],[I[i],S[i],beta,var,inclusion_length,skipping_length]);
        ##print('test2');#print(f1_2der);
        f1_der=(logit(psi[i])-logit(beta))/beta/(1-beta)/var;
        f1=I[i]*log(new_psi)+S[i]*log(1-new_psi)-pow(logit(psi[i])-logit(beta),2)/2/var+log(psi[i])+log(1-psi[i]);
        sum+=(0.5*f1_3der/(f1_2der)-f1_der);
    return(sum);

def myfunc_marginal_1(x, *args):
    beta2=x;beta1=x+cutoff;
    I1=args[0];S1=args[1];psi1=args[2];var1=args[3];
    I2=args[4];S2=args[5];psi2=args[6];var2=args[7];
    inclusion_length=args[8];skipping_length=args[9];
    return(myfunc_marginal(beta1,I1,S1,psi1,var1,inclusion_length,skipping_length)+myfunc_marginal(beta2,I2,S2,psi2,var2,inclusion_length,skipping_length));

def myfunc_marginal_1_der(x, *args):
    beta2=x;beta1=x+cutoff;
    I1=args[0];S1=args[1];psi1=args[2];var1=args[3];
    I2=args[4];S2=args[5];psi2=args[6];var2=args[7];
    inclusion_length=args[8];skipping_length=args[9];
    return(myfunc_marginal_der(beta1,I1,S1,psi1,var1,inclusion_length,skipping_length)+myfunc_marginal_der(beta2,I2,S2,psi2,var2,inclusion_length,skipping_length));

def myfunc_marginal_2(x, *args):
    beta2=x+cutoff;beta1=x;
    I1=args[0];S1=args[1];psi1=args[2];var1=args[3];
    I2=args[4];S2=args[5];psi2=args[6];var2=args[7];
    inclusion_length=args[8];skipping_length=args[9];
    return(myfunc_marginal(beta1,I1,S1,psi1,var1,inclusion_length,skipping_length)+myfunc_marginal(beta2,I2,S2,psi2,var2,inclusion_length,skipping_length));

def myfunc_marginal_2_der(x, *args):
    beta2=x+cutoff;beta1=x;
    I1=args[0];S1=args[1];psi1=args[2];var1=args[3];
    I2=args[4];S2=args[5];psi2=args[6];var2=args[7];
    inclusion_length=args[8];skipping_length=args[9];
    return(myfunc_marginal_der(beta1,I1,S1,psi1,var1,inclusion_length,skipping_length)+myfunc_marginal_der(beta2,I2,S2,psi2,var2,inclusion_length,skipping_length));
    
def myfunc_individual(x,*args):
    I=args[0];S=args[1];beta=args[2];var=args[3];
    inclusion_length=args[4];
    skipping_length=args[5]
    new_psi=inclusion_length*x/(inclusion_length*x+skipping_length*(1-x));
    return(-1*(I*log(new_psi)+S*log(1-new_psi)-(logit(x)-logit(beta))*(logit(x)-logit(beta))/2/var-log(x)-log(1-x)));

def myfunc_individual_der(x,*args):
    I=args[0];S=args[1];beta=args[2];var=args[3];
    inclusion_length=args[4];
    skipping_length=args[5];
    new_psi=inclusion_length*x/(inclusion_length*x+skipping_length*(1-x));
    new_psi_der=inclusion_length*skipping_length/pow(inclusion_length*x+skipping_length*(1-x),2);
    return(-1*(I/new_psi*new_psi_der-S/(1-new_psi)*new_psi_der-(logit(x)-logit(beta))/var/x/(1-x)-1/x+1/(1-x) ));

def myfunc_likelihood(x, args):
    I=args[0];S=args[1];beta=args[2];var=args[3];sum=0;N=I+S;
    #return(-1*(-log(sqrt((I+S)*x*(1-x)))-(I-(I+S)*x)*(I-(I+S)*x)/2/((I+S)*x*(1-x))-log(sqrt(var))-(x-beta)*(x-beta)/2/var));
    ##print('debug');#print(N);#print(var);#print(x);#print(beta);
    if N==0:
        return(0);
    return(-0.5*((I-N*x)*(I-N*x)/(N*x)+(S-N*(1-x))*(S-N*(1-x))/(N*(1-x)))-log(sqrt(var))-(x-beta)*(x-beta)/2/var);

def MLE_iteration_constrain(i1,i2,s1,s2,effective_inclusion_length,effective_skipping_length):
    psi1=vec2psi(i1,s1,effective_inclusion_length,effective_skipping_length);psi2=vec2psi(i2,s2,effective_inclusion_length,effective_skipping_length);
    iter_cutoff=1;iter_maxrun=100;count=0;previous_sum=0;
    beta_0=sum(psi1)/len(psi1);
    beta_1=sum(psi2)/len(psi2);
    var1=10*scipy.var(np.array(psi1)-beta_0);
    var2=10*scipy.var(np.array(psi2)-beta_1);
    if var1<=0.01:
        var1=0.01;
    if var2<=0.01:
        var2=0.01;
    #print('var1');#print(var1);#print('var2');#print(var2);
    while((iter_cutoff>0.01)&(count<=iter_maxrun)):
        count+=1;
        #iteration of beta
        beta_0=sum(psi1)/len(psi1);
        beta_1=sum(psi2)/len(psi2);
        #print('var1');#print(var1);#print('var2');#print(var2);
        #if abs(sum(psi1)/len(psi1)-sum(psi2)/len(psi2))>cutoff:
        if (sum(psi1)/len(psi1))>(sum(psi2)/len(psi2)):#minize psi2 if this is the case
            xopt = fmin_l_bfgs_b(myfunc_1,[sum(psi2)/len(psi2)],myfunc_der_1,args=[psi1,psi2,var1,var2],bounds=[[0.001,0.999-cutoff]],iprint=-1)
            theta2 = max(min(float(xopt[0]),1-cutoff),0);theta1=theta2+cutoff;
        else:#minize psi1 if this is the case
            xopt = fmin_l_bfgs_b(myfunc_2,[sum(psi1)/len(psi1)],myfunc_der_2,args=[psi1,psi2,var1,var2],bounds=[[0.001,0.999-cutoff]],iprint=-1)
            theta1 = max(min(float(xopt[0]),1-cutoff),0);theta2=theta1+cutoff;
        #print('constrain_1xopt');#print('theta');#print(theta1);#print(theta2);#print(xopt);
        #else:
        #    theta1=sum(psi1)/len(psi1);theta2=sum(psi2)/len(psi2);
        beta_0=theta1;beta_1=theta2;
        #iteration of psi
        new_psi1=[];new_psi2=[];current_sum=0;likelihood_sum=0;
        #print('constrain_2xopt');
        for i in range(len(psi1)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi1[i]],myfunc_individual_der,args=[i1[i],s1[i],beta_0,var1,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1);
            new_psi1.append(float(xopt[0]));current_sum+=float(xopt[1]);#print(xopt);
            #likelihood_sum+=myfunc_marginal(new_psi1[i],[i1[i],s1[i],beta_0,var1,effective_inclusion_length,effective_skipping_length]);
        for i in range(len(psi2)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi2[i]],myfunc_individual_der,args=[i2[i],s2[i],beta_1,var2,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1);
            new_psi2.append(float(xopt[0]));current_sum+=float(xopt[1]);#print(xopt);
            #likelihood_sum+=myfunc_marginal(new_psi2[i],[i2[i],s2[i],beta_1,var2,effective_inclusion_length,effective_skipping_length]);
        #print('new_psi[0]');#print(new_psi1[0]);#print(new_psi2[0]);
        psi1=new_psi1;psi2=new_psi2;
        #print('count');#print(count);#print('previous_sum');#print(previous_sum);#print('current_sum');#print(current_sum);
        if count>1:
            iter_cutoff=abs(previous_sum-current_sum);
        previous_sum=current_sum;
    ##print('constrain');#print(theta1);#print(theta2);#print(psi1);#print(psi2);#print(current_sum);#print(likelihood_sum);
    ##print(xopt);
    return([current_sum,[psi1,psi2,beta_0,beta_1,var1,var2]]);
    #return([likelihood_sum,[psi1,psi2,beta_0,beta_1,var1,var2]]);

def MLE_iteration(i1,i2,s1,s2,effective_inclusion_length,effective_skipping_length):
    psi1=vec2psi(i1,s1,effective_inclusion_length,effective_skipping_length);psi2=vec2psi(i2,s2,effective_inclusion_length,effective_skipping_length);
    iter_cutoff=1;iter_maxrun=100;count=0;previous_sum=0;
    beta_0=sum(psi1)/len(psi1);
    beta_1=sum(psi2)/len(psi2);
    var1=10*scipy.var(np.array(psi1)-beta_0);
    var2=10*scipy.var(np.array(psi2)-beta_1);
    if var1<=0.01:
        var1=0.01;
    if var2<=0.01:
        var2=0.01;
    #print('var1');#print(var1);#print('var2');#print(var2);
    while((iter_cutoff>0.01)&(count<=iter_maxrun)):
        count+=1;
        #iteration of beta
        beta_0=sum(psi1)/len(psi1);
        beta_1=sum(psi2)/len(psi2);
        xopt=fmin_l_bfgs_b(myfunc_multivar,[beta_0,beta_1],myfunc_multivar_der,args=[psi1,psi2,var1,var2],bounds=[[0.01,0.99],[0.01,0.99]],iprint=-1);
        beta_0=float(xopt[0][0]);
        beta_1=float(xopt[0][1]);
        #print('unconstrain_1xopt');#print(xopt);
        #print('theta');#print(beta_0);#print(beta_1);#print('theta_end');
        #iteration of psi
        new_psi1=[];new_psi2=[];current_sum=0;likelihood_sum=0;
        for i in range(len(psi1)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi1[i]],myfunc_individual_der,args=[i1[i],s1[i],beta_0,var1,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1);
            new_psi1.append(float(xopt[0]));current_sum+=float(xopt[1]);#print(xopt);
            #likelihood_sum+=myfunc_marginal(new_psi1[i],[i1[i],s1[i],beta_0,var1,effective_inclusion_length,effective_skipping_length]);
        for i in range(len(psi2)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi2[i]],myfunc_individual_der,args=[i2[i],s2[i],beta_1,var2,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1);
            new_psi2.append(float(xopt[0]));current_sum+=float(xopt[1]);#print(xopt);
            #likelihood_sum+=myfunc_marginal(new_psi2[i],[i2[i],s2[i],beta_1,var2,effective_inclusion_length,effective_skipping_length]);
        #print('new_psi[0]');#print(new_psi1[0]);#print(new_psi2[0]);
        psi1=new_psi1;psi2=new_psi2;print
        #print('count');#print(count);('previous_sum');#print(previous_sum);#print('current_sum');#print(current_sum);
        if count>1:
            iter_cutoff=abs(previous_sum-current_sum);
        previous_sum=current_sum;
    ##print('unconstrain');#print(beta_0);#print(beta_0+beta_1);#print(psi1);#print(psi2);#print(current_sum);#print(likelihood_sum);
    ##print(xopt);
    if count>iter_maxrun:
        return([current_sum,[psi1,psi2,0,0,var1,var2]]);
    return([current_sum,[psi1,psi2,beta_0,beta_1,var1,var2]]);
    #return([likelihood_sum,[psi1,psi2,beta_0,beta_1,var1,var2]]);

def MLE_marginal_iteration(i1,i2,s1,s2,effective_inclusion_length,effective_skipping_length):
    #initial value
    psi1=vec2psi(i1,s1,effective_inclusion_length,effective_skipping_length);psi2=vec2psi(i2,s2,effective_inclusion_length,effective_skipping_length);
    iter_cutoff=1;iter_maxrun=100;count=0;previous_sum=0;
    beta_0=sum(psi1)/len(psi1);
    beta_1=sum(psi2)/len(psi2);
    #var1=scipy.var(np.array(veclogit(psi1)));
    #var2=scipy.var(np.array(veclogit(psi2)));
    var1=scipy.var(np.array(psi1)-beta_0);
    var2=scipy.var(np.array(psi2)-beta_1);
    if var1<=0.01:
        var1=0.01;
    if var2<=0.01:
        var2=0.01;
    #MLE of the full likelihood
    while((iter_cutoff>0.01)&(count<=iter_maxrun)):
        count+=1;
        beta_0=sum(psi1)/len(psi1);
        beta_1=sum(psi2)/len(psi2);
        xopt=fmin_l_bfgs_b(myfunc_multivar,[beta_0,beta_1],myfunc_multivar_der,args=[psi1,psi2,var1,var2],bounds=[[0.01,0.99],[0.01,0.99]],iprint=-1);
        beta_0=float(xopt[0][0]);
        beta_1=float(xopt[0][1]);
        new_psi1=[];new_psi2=[];current_sum=0;likelihood_sum=0;
        for i in range(len(psi1)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi1[i]],myfunc_individual_der,args=[i1[i],s1[i],beta_0,var1,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1);
            new_psi1.append(float(xopt[0]));current_sum+=float(xopt[1]);
        for i in range(len(psi2)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi2[i]],myfunc_individual_der,args=[i2[i],s2[i],beta_1,var2,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1);
            new_psi2.append(float(xopt[0]));current_sum+=float(xopt[1]);
        psi1=new_psi1;psi2=new_psi2;
        if count>1:
            iter_cutoff=abs(previous_sum-current_sum);
        previous_sum=current_sum;
    if count>iter_maxrun:
        psi1=vec2psi(i1,s1,effective_inclusion_length,effective_skipping_length);psi2=vec2psi(i2,s2,effective_inclusion_length,effective_skipping_length);
        iter_cutoff=1;iter_maxrun=100;count=0;previous_sum=0;
        beta_0=sum(psi1)/len(psi1);
        beta_1=sum(psi2)/len(psi2);
        #var1=scipy.var(np.array(veclogit(psi1)));
        #var2=scipy.var(np.array(veclogit(psi2)));
        var1=10*scipy.var(np.array(psi1)-beta_0);
        var2=10*scipy.var(np.array(psi2)-beta_1);
        if var1<=0.01:
            var1=0.01;
        if var2<=0.01:
            var2=0.01;
        while((iter_cutoff>0.01)&(count<=iter_maxrun)):
            count+=1;
            beta_0=sum(psi1)/len(psi1);
            beta_1=sum(psi2)/len(psi2);
            xopt=fmin_l_bfgs_b(myfunc_multivar,[beta_0,beta_1],myfunc_multivar_der,args=[psi1,psi2,var1,var2],bounds=[[0.01,0.99],[0.01,0.99]],iprint=-1);
            beta_0=float(xopt[0][0]);
            beta_1=float(xopt[0][1]);
            new_psi1=[];new_psi2=[];current_sum=0;likelihood_sum=0;
            for i in range(len(psi1)):
                xopt = fmin_l_bfgs_b(myfunc_individual,[psi1[i]],myfunc_individual_der,args=[i1[i],s1[i],beta_0,var1,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1);
                new_psi1.append(float(xopt[0]));current_sum+=float(xopt[1]);
            for i in range(len(psi2)):
                xopt = fmin_l_bfgs_b(myfunc_individual,[psi2[i]],myfunc_individual_der,args=[i2[i],s2[i],beta_1,var2,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1);
                new_psi2.append(float(xopt[0]));current_sum+=float(xopt[1]);
            psi1=new_psi1;psi2=new_psi2;
            if count>1:
                iter_cutoff=abs(previous_sum-current_sum);
            previous_sum=current_sum;
            if count>iter_maxrun:
                return([current_sum,[psi1,psi2,0,0,var1,var2,0]]);
    
    #MLE of the marginal likelihood
    iter_cutoff=1;iter_maxrun=100;count=0;previous_sum=0;
    while((iter_cutoff>0.01)&(count<=iter_maxrun)):
        count+=1;
        ##print('unconstrain_MLE_marginal_value_1');#print(myfunc_marginal(0.51,i1,s1,psi1,var1,effective_inclusion_length,effective_skipping_length));
        ##print('unconstrain_MLE_marginal_value_2');#print(myfunc_marginal(0.56,i2,s2,psi2,var2,effective_inclusion_length,effective_skipping_length));
        #exit();
        xopt1=fmin_l_bfgs_b(myfunc_marginal,[beta_0],myfunc_marginal_der,args=[i1,s1,psi1,var1,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1);
        xopt2=fmin_l_bfgs_b(myfunc_marginal,[beta_1],myfunc_marginal_der,args=[i2,s2,psi2,var2,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1);
        beta_0=float(xopt1[0][0]);
        beta_1=float(xopt2[0][0]);
        #print(xopt1)
        ##if re.findall('\d+',str(xopt1[2]))[0]!='0':
        #    return([current_sum,[psi1,psi2,0,0,var1,var2,0]]);
        ##if re.findall('\d+',str(xopt2[2]))[0]!='0':
        #    return([current_sum,[psi1,psi2,0,0,var1,var2,0]]);
        #print('unconstrain_1xopt');print(xopt1);
        #print('unconstrain_2xopt');print(xopt2);
        new_psi1=[];new_psi2=[];current_sum=0;likelihood_sum=0;
        for i in range(len(psi1)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi1[i]],myfunc_individual_der,args=[i1[i],s1[i],beta_0,var1,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1);
            new_psi1.append(float(xopt[0]));current_sum+=float(xopt[1]);
        for i in range(len(psi2)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi2[i]],myfunc_individual_der,args=[i2[i],s2[i],beta_1,var2,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1);
            new_psi2.append(float(xopt[0]));current_sum+=float(xopt[1]);
        ##if re.findall('\d+',str(xopt[2]))[0]!='0':
            #return([current_sum,[psi1,psi2,0,0,var1,var2,0]]);
        psi1=new_psi1;psi2=new_psi2;
        if count>1:
            iter_cutoff=abs(previous_sum-current_sum);
        previous_sum=current_sum;
    cor=max(min(stats.pearsonr(psi1,psi2)[0],0.9),-0.9);
    #print(cor)
    if math.isnan(cor):
        cor=0;
    if count>iter_maxrun:
        return([current_sum,[psi1,psi2,0,0,var1,var2,cor]]);
    return([current_sum,[psi1,psi2,beta_0,beta_1,var1,var2,cor]]);

def MLE_marginal_iteration_constrain(i1,i2,s1,s2,effective_inclusion_length,effective_skipping_length):
    #initial value
    psi1=vec2psi(i1,s1,effective_inclusion_length,effective_skipping_length);psi2=vec2psi(i2,s2,effective_inclusion_length,effective_skipping_length);
    iter_cutoff=1;iter_maxrun=100;count=0;previous_sum=0;
    beta_0=sum(psi1)/len(psi1);
    beta_1=sum(psi2)/len(psi2);
    #var1=scipy.var(np.array(veclogit(psi1)));
    #var2=scipy.var(np.array(veclogit(psi2)));
    var1=scipy.var(np.array(psi1)-beta_0);
    var2=scipy.var(np.array(psi2)-beta_1);
    if var1<=0.01:
        var1=0.01;
    if var2<=0.01:
        var2=0.01;
    #MLE of the full likelihood
    while((iter_cutoff>0.01)&(count<=iter_maxrun)):
        count+=1;
        beta_0=sum(psi1)/len(psi1);
        beta_1=sum(psi2)/len(psi2);
        if (sum(psi1)/len(psi1))>(sum(psi2)/len(psi2)):#minize psi2 if this is the case
            xopt = fmin_l_bfgs_b(myfunc_1,[sum(psi2)/len(psi2)],myfunc_der_1,args=[psi1,psi2,var1,var2],bounds=[[0.001,0.999-cutoff]],iprint=-1)
            theta2 = max(min(float(xopt[0]),1-cutoff),0);theta1=theta2+cutoff;
        else:#minize psi1 if this is the case
            xopt = fmin_l_bfgs_b(myfunc_2,[sum(psi1)/len(psi1)],myfunc_der_2,args=[psi1,psi2,var1,var2],bounds=[[0.001,0.999-cutoff]],iprint=-1)
            theta1 = max(min(float(xopt[0]),1-cutoff),0);theta2=theta1+cutoff;
        beta_0=theta1;beta_1=theta2;
        new_psi1=[];new_psi2=[];current_sum=0;likelihood_sum=0;
        for i in range(len(psi1)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi1[i]],myfunc_individual_der,args=[i1[i],s1[i],beta_0,var1,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1);
            new_psi1.append(float(xopt[0]));current_sum+=float(xopt[1]);#print(xopt);
        for i in range(len(psi2)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi2[i]],myfunc_individual_der,args=[i2[i],s2[i],beta_1,var2,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1);
            new_psi2.append(float(xopt[0]));current_sum+=float(xopt[1]);#print(xopt);
        psi1=new_psi1;psi2=new_psi2;
        if count>1:
            iter_cutoff=abs(previous_sum-current_sum);
        previous_sum=current_sum;
    if count>iter_maxrun:
        psi1=vec2psi(i1,s1,effective_inclusion_length,effective_skipping_length);psi2=vec2psi(i2,s2,effective_inclusion_length,effective_skipping_length);
        iter_cutoff=1;iter_maxrun=100;count=0;previous_sum=0;
        beta_0=sum(psi1)/len(psi1);
        beta_1=sum(psi2)/len(psi2);
        #var1=scipy.var(np.array(veclogit(psi1)));
        #var2=scipy.var(np.array(veclogit(psi2)));
        var1=10*scipy.var(np.array(psi1)-beta_0);
        var2=10*scipy.var(np.array(psi2)-beta_1);
        if var1<=0.01:
            var1=0.01;
        if var2<=0.01:
            var2=0.01;
        #MLE of the full likelihood
        while((iter_cutoff>0.01)&(count<=iter_maxrun)):
            count+=1;
            beta_0=sum(psi1)/len(psi1);
            beta_1=sum(psi2)/len(psi2);
            if (sum(psi1)/len(psi1))>(sum(psi2)/len(psi2)):#minize psi2 if this is the case
                xopt = fmin_l_bfgs_b(myfunc_1,[sum(psi2)/len(psi2)],myfunc_der_1,args=[psi1,psi2,var1,var2],bounds=[[0.001,0.999-cutoff]],iprint=-1)
                theta2 = max(min(float(xopt[0]),1-cutoff),0);theta1=theta2+cutoff;
            else:#minize psi1 if this is the case
                xopt = fmin_l_bfgs_b(myfunc_2,[sum(psi1)/len(psi1)],myfunc_der_2,args=[psi1,psi2,var1,var2],bounds=[[0.001,0.999-cutoff]],iprint=-1)
                theta1 = max(min(float(xopt[0]),1-cutoff),0);theta2=theta1+cutoff;
            beta_0=theta1;beta_1=theta2;
            new_psi1=[];new_psi2=[];current_sum=0;likelihood_sum=0;
            for i in range(len(psi1)):
                xopt = fmin_l_bfgs_b(myfunc_individual,[psi1[i]],myfunc_individual_der,args=[i1[i],s1[i],beta_0,var1,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1);
                new_psi1.append(float(xopt[0]));current_sum+=float(xopt[1]);#print(xopt);
            for i in range(len(psi2)):
                xopt = fmin_l_bfgs_b(myfunc_individual,[psi2[i]],myfunc_individual_der,args=[i2[i],s2[i],beta_1,var2,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1);
                new_psi2.append(float(xopt[0]));current_sum+=float(xopt[1]);#print(xopt);
            psi1=new_psi1;psi2=new_psi2;
            if count>1:
                iter_cutoff=abs(previous_sum-current_sum);
            previous_sum=current_sum;
        if count>iter_maxrun:
            return([current_sum,[psi1,psi2,0,0,var1,var2]]);

    
    #MLE of the marginal likelihood
    iter_cutoff=1;iter_maxrun=100;count=0;previous_sum=0;
    while((iter_cutoff>0.01)&(count<=iter_maxrun)):
        count+=1;
        if (sum(psi1)/len(psi1))>(sum(psi2)/len(psi2)):#minize psi2 if this is the case
            xopt = fmin_l_bfgs_b(myfunc_marginal_1,[beta_1],myfunc_marginal_1_der,args=[i1,s1,psi1,var1,i2,s2,psi2,var2,effective_inclusion_length,effective_skipping_length],bounds=[[0.001,0.999-cutoff]],iprint=-1)
            beta_1 = max(min(float(xopt[0]),1-cutoff),0);beta_0=beta_1+cutoff;
        else:#minize psi1 if this is the case
            xopt = fmin_l_bfgs_b(myfunc_marginal_2,[beta_0],myfunc_marginal_2_der,args=[i1,s1,psi1,var1,i2,s2,psi2,var2,effective_inclusion_length,effective_skipping_length],bounds=[[0.001,0.999-cutoff]],iprint=-1)
            beta_0 = max(min(float(xopt[0]),1-cutoff),0);beta_1=beta_0+cutoff;
        #if re.findall('\d+',str(xopt[2]))[0]!='0':
            #return([current_sum,[psi1,psi2,0,0,var1,var2]]);
        #print('constrain_xopt');#print(xopt);
        new_psi1=[];new_psi2=[];current_sum=0;likelihood_sum=0;
        for i in range(len(psi1)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi1[i]],myfunc_individual_der,args=[i1[i],s1[i],beta_0,var1,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1);
            new_psi1.append(float(xopt[0]));current_sum+=float(xopt[1]);#print(xopt);
        for i in range(len(psi2)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi2[i]],myfunc_individual_der,args=[i2[i],s2[i],beta_1,var2,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1);
            new_psi2.append(float(xopt[0]));current_sum+=float(xopt[1]);#print(xopt);
        #if re.findall('\d+',str(xopt[2]))[0]!='0':
            #return([current_sum,[psi1,psi2,0,0,var1,var2]]);
        psi1=new_psi1;psi2=new_psi2;
        if count>1:
            iter_cutoff=abs(previous_sum-current_sum);
        previous_sum=current_sum;
    if count>iter_maxrun:
        return([current_sum,[psi1,psi2,0,0,var1,var2]]);

    return([current_sum,[psi1,psi2,beta_0,beta_1,var1,var2]]);
    #return([likelihood_sum,[psi1,psi2,beta_0,beta_1,var1,var2]]);
    


    
#Random Sampling Function
def likelihood_test(i1,i2,s1,s2,effective_inclusion_length,effective_skipping_length,flag,id):
    #print('testing'+str(id));
    if flag==0:
        return([1,0]);
    else:
        res=MLE_marginal_iteration(i1,i2,s1,s2,effective_inclusion_length,effective_skipping_length);
        #print(res)
        if abs(res[1][2]-res[1][3])<=cutoff:
            #print('1<=cutoff');#print(res);#print((res[1][2]-res[1][3]));
            #return([1-scipy.stats.norm.cdf(abs(res[1][2]-res[1][3]),cutoff,sqrt(res[1][4]+res[1][5])),res[0]]);
            return([1,0]);
        else:
            res_constrain=MLE_marginal_iteration_constrain(i1,i2,s1,s2,effective_inclusion_length,effective_skipping_length);
            #print('2>cutoff');#print(res);#print(res_constrain);
            #print(abs(res_constrain[0]-res[0]));#print(pow(1-pow(res[1][6],2),2));#print('2end');
            ##print(abs(res_constrain[2]-res[2]));#print('2end_marginal');            
            return([scipy.stats.chi2.sf(2*(abs(res_constrain[0]-res[0]))/pow(1-pow(res[1][6],2),2),1),res[1][-1]]);

#MultiProcessorFunction
def MultiProcessorPool(n_original_diff):
    i1=n_original_diff[0];i2=n_original_diff[1];s1=n_original_diff[2];s2=n_original_diff[3];
    effective_inclusion_length=n_original_diff[4];effective_skipping_length=n_original_diff[5];
    flag=n_original_diff[6];id=n_original_diff[7];
    P=likelihood_test(i1,i2,s1,s2,effective_inclusion_length,effective_skipping_length,flag,id);
    return(P);

#Function for vector handling
def vec2float(vec):
    res=[];
    for i in vec:
        res.append(float(i));
    return(res);

#add 1 in both inclusion and skipping counts for robustness in small sample size
def vecAddOne(vec):
    res=[];
    for i in vec:
        res.append(i+1);
    return(res);

def vecprod(vec):
    res=1;
    for i in vec:
        res=res*i;
    return(res);

def vecadd(vec1,vec2):
    res=[];
    for i in range(len(vec1)):
        res.append(vec1[i]+vec2[i]);
    return(res);
    
def vec2remove0psi(inc,skp):
    res1=[];res2=[];
    for i in range(len(inc)):
        if (inc[i]!=0) | (skp[i]!=0):
            res1.append(inc[i]);res2.append(skp[i]);
    return([res1,res2]);

def vec2psi(inc,skp,effective_inclusion_length,effective_skipping_length):
    psi=[];
    inclusion_length=effective_inclusion_length;
    skipping_length=effective_skipping_length;
    for i in range(len(inc)):
        if (float(inc[i])+float(skp[i]))==0:
            psi.append(0.5);
        else:
            psi.append(float(inc[i])/inclusion_length/(float(inc[i])/inclusion_length+float(skp[i])/skipping_length));
    return(psi);

def vec210(vec):
    res=[];
    for i in vec:
        if i>0:
            res.append(1);
        else:
            res.append(-1);
    return(res);

#this function removes the paired samples with inc1+skp1=0 or inc2+skp2=0
def vec_remove_na_pair(inc1,skp1,inc2,skp2):
    res_inc1=[];res_skp1=[];res_inc2=[];res_skp2=[];
    for i in range(len(inc1)):
        if ((float(inc1[i])+float(skp1[i]))>0) & ((float(inc2[i])+float(skp2[i]))>0):
            res_inc1.append(inc1[i]);
            res_skp1.append(skp1[i]);
            res_inc2.append(inc2[i]);
            res_skp2.append(skp2[i]);
    return([res_inc1,res_skp1,res_inc2,res_skp2]);
    
def vec_remove_na(inc,skp):
    res_inc=[];res_skp=[];
    for i in range(len(inc)):
        if (float(inc[i])+float(skp[i]))>0:
            res_inc.append(inc[i]);
            res_skp.append(skp[i]);
    return([res_inc,res_skp]);
def myorder(p,reverse):
    p_dict={}
    ##print('len(p)');#print(len(p))
    for index in range(len(p)):
        if not(p[index] in p_dict):
            p_dict[p[index]]=[]
        p_dict[p[index]].append(index+1)
    ##print('p_dict');#print(p_dict)
    o_index=sorted(p_dict,reverse=reverse);o=[]
    ##print('o_index');#print(o_index)
    for index in o_index:
        for this_p in p_dict[index]:
            o.append(this_p)
    return(o)
def mycummin(p):
        res=[]
        for i in range(len(p)):
            res.append(min(p[:(i+1)]))
        return(res)

def myFDR(p):
    lp=len(p)
    i=range(lp,0,-1)
    o=myorder(p,True)
    ro=myorder(o,False)
    p_new=[]
    for index in range(len(o)):
        p_new.append(p[o[index]-1]*lp/i[index])
    p_mycummin=mycummin(p_new)
    p_mycummin_new=[]
    for index in range(len(p_mycummin)):
        p_mycummin_new.append(min(p_mycummin[index],1))
    res=[]
    for index in range(len(p_mycummin_new)):
        res.append(p_mycummin_new[ro[index]-1])
    return(res)

def adjustPsi(inc,skp,effective_inclusion_length,effective_skipping_length):
    #if effective_inclusion_length == 0:
        #print([inc,skp,effective_inclusion_length,effective_skipping_length])
    inc_adj=np.array(inc)/effective_inclusion_length
    skp_adj=np.array(skp)/effective_skipping_length
    return(','.join([str(i) for i in (inc_adj/(inc_adj+skp_adj)).tolist()]))

def myttest(vec1,vec2):
    if (len(vec1)==1) & (len(vec2)==1):
        res=stats.ttest_ind([vec1[0],vec1[0]],[vec2[0],vec2[0]]);
    else:
        res=stats.ttest_ind(vec1,vec2);
    return(res);

def myand(l1,l2):
    return([i for i in range(len(l1)) if l1[i] & l2[i] ==True])

    ######## end of functions ##############
def getrMATS_P(list_n_original_diff,mycutoff,threads):
    global cutoff
    cutoff=mycutoff
    probability=[]
    if threads>1:
        pool=Pool(processes=threads)
        probability=pool.map(MultiProcessorPool,list_n_original_diff)
        pool.close()
        pool.join()
    else:
        for i in range(len(list_n_original_diff)):
            probability.append(MultiProcessorPool(list_n_original_diff[i]))
    P=pd.DataFrame(probability).iloc[:,0].tolist()
    return(P)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author: Zelin Liu
email: zlliu@bjmu.edu.cn
license: GPL3
detail: Analysis differential back-spliced in of rRNA-depleted RNA-seq
"""
from docopt import docopt
import sys
from .version import __version__
#__version__=1.01

helpInfo = '''
Usage: DEBKS <command> [options]

Command:
    merge            Merge circRNA junction from other software
    count            Count linear junction with BAM file
    anno             Annotate circRNA with gene annotation
    dec              Detect differentially expressed circRNA with junction information
'''

def main():
    command_log = 'DEBKS parameters: ' + ' '.join(sys.argv)
    if len(sys.argv) == 1:
        sys.exit(helpInfo)
    elif sys.argv[1] == '--version' or sys.argv[1] == '-v':
        sys.exit(__version__)
    elif sys.argv[1] == 'dec':
        from . import dec
        dec.dec(docopt(dec.__doc__, version=__version__))
    elif sys.argv[1] == 'merge':
        from . import merge
        merge.merge(docopt(merge.__doc__, version=__version__))
    elif sys.argv[1] == 'anno':
        from . import anno
        anno.anno(docopt(anno.__doc__, version=__version__))
    elif sys.argv[1] == 'count':
        from . import count
        count.count(docopt(count.__doc__, version=__version__))
    else:
        sys.exit(helpInfo)

if __name__ == '__main__':
    main()



'''
Usage: DEBKS  merge (-s software | -d designate) (-n name | -f file)  [-b pos_based]  [-o output]

Options:
    -h --help                   Show help message.
    -v --version                Show version.
    -s software                 Name of software used to detect circRNA (e.g. ciri2, circexplorer2, find_circ).
    -d designate                Designate the column postion of chr,start,end,circ,linear in the output file of detection software (e.g. 1,2,3,4,5).
                                The first four postion is required.
    -b pos_based                0- or 1-based position for designated start and end position, separated by comma [default: 1,1].
    -n name                     Names of output file of detection software, separated by comma (e.g. s1,s2,s3).
    -f file                     File includes names of output file of detection software, separated by line breaks.
    -o output                   Output prefix [default: merge].
'''
import pysam,pandas as pd, numpy as np,time,os,sys,traceback

def fileCheck(file):
    for i in file:
        if not os.path.exists(i):
            sys.exit('ERROR: %s is not exist!!!' % i)

def checkD(x):
    x=x.split(',')
    try:
        arr=[int(i) for i in x]
    except:
        sys.exit('ERROR: make sure "-d" options is suitable!!!')
    if len(arr) not in [4,5]:
        sys.exit('ERROR: "-d" options required four or five positions!!!')
    if 0 in arr:
        sys.exit('ERROR: "-d" options is 1-based, 0 position is not allowed!!!')
    return(arr)

def adjPos(pos_mat,based):
    based=based.split(',')
    if len(based)!=2:
        sys.exit('ERROR: "-b" only accept two parameter separated by comma!!!')
    if based[0]=='0':
        pos_mat['start']=pos_mat['start'].map(lambda x:int(x)+1).tolist()
    if based[1]=='0':
        pos_mat['start']=pos_mat['start'].map(lambda x:int(x)+1).tolist()
    return(pos_mat)
def merge(options):
    output=options['-o']
    soft_all=['ciri2','circexplorer2','find_circ']
    pos_ciri2=[2,3,4,5,7] # 1-based
    pos_circexplorer2=[1,2,3,4] # start: 0-based  end: 1-based
    pos_find_circ=[1,2,3,7] # Both start and end are 0-based
    pos_dict={'ciri2':pos_ciri2,'circexplorer2':pos_circexplorer2,'find_circ':pos_find_circ}
    if options['-s']:
        software=options['-s'].lower()
        if software not in soft_all:
            sys.exit('ERROR: %s is not included in our default software list, please try "-d" to designate the position' % options['-s'])
        circ_pos=pos_dict[software]
    else:
        circ_pos=checkD(options['-d'])
    circ_pos=[i-1 for i in circ_pos]
    # read file
    fileArr=[]
    if options['-n']:
        fileArr=options['-n'].split(',')
    else:
        if os.path.exists(options['-f']):
            fileArr=[i.strip() for i in open(options['-f']).readlines()]
        else:
            sys.exit('ERROR: %s is not exist!!!' % options['-f'])
    fileArr_2=list(set(fileArr))
    if len(fileArr_2)<len(fileArr):
        sys.exit('ERROR: file names is replicated!!!')
    fileCheck(fileArr)
    stime=time.time()
    print('Merge %d files.' % len(fileArr))
    if len(circ_pos)<5:
        pos_mat,circ_mat=comMat(fileArr,circ_pos)
    else:
        pos_mat,circ_mat,linear_mat=comMat(fileArr,circ_pos)
    if options['-s']:
        if software=='circexplorer2':
            pos_mat['start']=pos_mat['start'].map(lambda x:int(x)+1).tolist()
        elif software=='find_circ':
            pos_mat['start']=pos_mat['start'].map(lambda x:int(x)+1).tolist()
            pos_mat['end']=pos_mat['end'].map(lambda x:int(x)+1).tolist()
    if options['-d'] and options['-b']:
        pos_mat=adjPos(pos_mat,options['-b'])
    pos_mat.to_csv(output+'_pos.txt',sep='\t',header=True,index=None)
    circ_mat=pd.concat([pos_mat,circ_mat],axis=1)
    circ_mat.to_csv(output+'_circ.txt',sep='\t',header=True,index=None)
    if len(circ_pos)==5:
        linear_mat=pd.concat([pos_mat,linear_mat],axis=1)
        linear_mat.to_csv(output+'_linear.txt',sep='\t',header=True,index=None)
    etime=time.time()-stime
    print('Finished in %d second.' % etime)

#### function part ####
def readMatFile(fileName,pos):
    fff=open(fileName)
    fl=fff.readline().strip().split('\t')
    f2=fff.readline().strip().split('\t')
    fff.close()
    if len(f2) <(max(pos)+1):
        sys.exit('ERROR: The number of columns in %s is not enough!!!' % fileName)
    isHeader=None
    for i in fl[pos[1]]:
        if i not in ['0','1','2','3','4','5','6','7','8','9']:
            isHeader=1
            break
    if isHeader:
        mat=pd.read_csv(fileName,sep='\t',skiprows=[0],header=None).iloc[:,pos]
    else:
        mat=pd.read_csv(fileName,sep='\t',header=None).iloc[:,pos]
    return(mat)

def comMat(fileArr,pos):
    circ_dict={}
    linear_dict={}
    for i in fileArr:
        tmpMat=readMatFile(i,pos)
        for j in range(tmpMat.shape[0]):
            tmp=tmpMat.iloc[j,:].tolist()
            circID=str(tmp[0])+'|'+str(tmp[1])+'|'+str(tmp[2])
            if not circ_dict.__contains__(circID):
                circ_dict[circID]={}
            circ_dict[circID][i]=tmp[3]
            if len(pos)==5:
                if not linear_dict.__contains__(circID):
                    linear_dict[circID]={}
                linear_dict[circID][i]=tmp[4]
    circ_mat={}
    linear_mat={}
    pos_mat=[]
    for circID in circ_dict.keys():
        pos_mat.append(circID.split('|'))
        circ_mat[circID]=[]
        linear_mat[circID]=[]
        for i in fileArr:
            if circ_dict[circID].__contains__(i):
                circ_mat[circID].append(circ_dict[circID][i])
            else:
                circ_mat[circID].append(0)
            if len(pos)==5:
                if linear_dict[circID].__contains__(i):
                    linear_mat[circID].append(linear_dict[circID][i])
                else:
                    linear_mat[circID].append(0)
    circ_mat=list(circ_mat.values())
    circ_mat=pd.DataFrame(circ_mat,columns=fileArr)
    pos_mat=pd.DataFrame(pos_mat,columns=['chr','start','end'])
    #circ_mat=pd.concat([pos_mat,circ_mat],axis=1)
    if len(pos)<5:
        return(pos_mat,circ_mat)
    else:
        linear_mat=list(linear_mat.values())
        linear_mat=pd.DataFrame(linear_mat,columns=fileArr)
        #linear_mat=pd.concat([pos_mat,linear_mat],axis=1)
        return(pos_mat,circ_mat,linear_mat)


'''
Usage: DEBKS dec -c circ (-l linear |--c2 circ2) [-p] [-n sample_num] [-f circRNA_filter] [-s sample_filter] [-t threads] [-d cutoff] [-r read_len] [-a hangover_len] [-e circ_len] [--e2 circ2_len] [-o output]

Options:
    -h --help                   Show help message.
    -v --version                Show version.
    -c circ                     Circular junction counts file in tab format, the first three columns is circRNA position (chr,start,end).
    -l linear                   Linear junction counts file, the format is same with -c.
    --c2 circ2                  Circular junction counts file in tab format, the format is same with -c.
    -p                          Sample group is paired.
    -n sample_num               Number of samples in each group, separated by comma (e.g. 2,3).
    -f circRNA_filter           Required total circular juction counts in all samples to filter out low expressed circRNAs [default: 0].
    -s sample_filter            Required lowest CJC of each circRNA to filter out sample [default: 0].
    -t threads                  Number of threads to detect DE circRNA [default: 4].
    -d cutoff                   Cutoff of DE circRNA in two groups [default: 0.05].
    -r read_len                 Length of RNA-seq reads [default: 150].
    -a hangover_len             The minimum hangover length for junction quantification [default: 6].
    -e circ_len                 File of circRNA length in tab format. The first four columns is circRNA position (chr, start, end) and circRNA length.
    --e2 circ2_len              File of circRNA2 length in tab format, only work with --c2. The format is same with -e.
    -o output                   Output file [default: dec_circRNA.txt].
'''
import sys,time,pandas as pd,numpy as np,docopt,os,time

def readJuncFile(fileName,reqLen=7):
    if not os.path.exists(fileName):
        sys.exit('ERROR: %s is not exist!!!' % fileName)
    fff=open(fileName)
    fl=fff.readline().strip().split('\t')
    fff.close()
    if len(fl) <reqLen:
        sys.exit('ERROR: The number of columns in %s is not enough!!!' % fileName)
    isHeader=None
    for i in fl[1]:
        if i not in ['0','1','2','3','4','5','6','7','8','9']:
            isHeader=0
            break
    juncMat=pd.read_csv(fileName,sep='\t',header=isHeader)
    return(juncMat)

def filterCirc(mat,filterNum):
    rowSum=mat.apply(lambda x: sum(x[3:]),axis=1).tolist()
    passID=[]
    for i in range(mat.shape[0]):
        if rowSum[i]>=filterNum:
            passID.append(i)
    return(passID)
    
def getCircLen(lenFile,circID,adjustReadLength):
    circLenFile=readJuncFile(lenFile,reqLen=4)
    circID3=(circLenFile.iloc[:,0]+':'+circLenFile.iloc[:,1].map(str)+'|'+circLenFile.iloc[:,2].map(str)).tolist()
    circLen=circLenFile.iloc[:,3].map(int).tolist()
    if len(circLen) != len(circID):
        sys.exit('ERROR: The number of lines in %s file (%d) is different with junction file (%d)!!!' % (lenFile,len(circLen),len(circID)))
    if circID !=circID3:
        sys.exit('ERROR: circRNA position is different in junction file and %s, DEBKS do not adjust order automatically!!!' % lenFile)
    if min(circLen)==0:
        sys.exit('ERROR: some circRNAs are designeted to 0 nt length in %s!!!' % lenFile)
    circLen=[int(i) for i in circLen]
    adjust_inc_len=[]
    adjust_skp_len=[]
    for i in circLen:
        if i<2*adjustReadLength:
            adjust_inc_len.append(2*adjustReadLength+i)
            adjust_skp_len.append(i)
        else:
            adjust_inc_len.append(4*adjustReadLength)
            adjust_skp_len.append(2*adjustReadLength)
    return(adjust_inc_len,adjust_skp_len)

def dec(options):
    #print(options)
    # number parameter
    threads=int(options['-t'])
    cutoff=min(float(options['-d']),1.00)
    readLength=int(options['-r'])
    hangOver=int(options['-a'])
    filterNum=max(int(options['-f']),0)
    filterSample=max(int(options['-s']),0)
    # is paired
    if options['-p']:
        from .rMATs_paired import getrMATS_P,vec2float,vecAddOne,adjustPsi,myFDR
        groupPair=True
    else:
        from .rMATs import getrMATS_P,vec2float,vecAddOne,adjustPsi,myFDR
        groupPair=False
    global getrMATS_P,vec2float,vecAddOne,adjustPsi,myFDR
    # junction file
    circFile=options['-c']
    circMat=readJuncFile(circFile)
    if options['-l']:
        linearFile=options['-l']
        linearMat=readJuncFile(linearFile)
    if options['--c2']:
        circ2File=options['--c2']
        linearMat=readJuncFile(circ2File)

    if circMat.shape[0]==linearMat.shape[0]:
        circID=(circMat.iloc[:,0]+':'+circMat.iloc[:,1].map(str)+'|'+circMat.iloc[:,2].map(str)).tolist()
        circID2=(linearMat.iloc[:,0]+':'+linearMat.iloc[:,1].map(str)+'|'+linearMat.iloc[:,2].map(str)).tolist()
        if not circID==circID2:
            sys.exit('ERROR: circRNA position is different in two junction file!!!')
    else:
        sys.exit('ERROR: The number of circRNA is different in two junction file!!!')
    # sample number
    sampleNum=circMat.shape[1]-3
    if options['-n']:
        group_num=options['-n'].split(',')
        group1_num=int(group_num[0])
        group2_num=int(group_num[1])
        if group1_num+group2_num !=sampleNum:
            sys.exit('ERROR: Sample number mismatch the junction file!!!')
    else:
        group1_num=int((circMat.shape[1]-3)/2)
        group2_num=sampleNum-group1_num
    if group1_num<2 or group2_num<2:
        sys.exit('ERROR: Require replicates!!!')
    # adjust inclusion and exclusion length
    adjustReadLength=readLength-hangOver
    if options['-e']:
        adjust_inc_len,adjust_skp_len=getCircLen(options['-e'],circID,adjustReadLength)
    else:
        adjust_inc_len=[4*adjustReadLength for i in range(circMat.shape[0])]
        adjust_skp_len=[2*adjustReadLength for i in range(circMat.shape[0])]
    if options['--c2']:
        if options['--e2']:
            tmp,adjust_inc_len=getCircLen(options['--e2'],circID2,adjustReadLength)
        else:
            adjust_inc_len=[2*adjustReadLength for i in range(circMat.shape[0])]
    
    if filterNum>0:
        passID=filterCirc(circMat,filterNum)
        if len(passID)>0:
            print('Keep %d circRNAs for dec.' % len(passID))
            circMat=circMat.iloc[passID,:].copy()
            linearMat=linearMat.iloc[passID,:].copy()
            adjust_inc_len=[adjust_inc_len[i] for i in passID]
            adjust_skp_len=[adjust_skp_len[i] for i in passID]
        else:
            sys.exit('WARNING: Filter keep zero circRNA for dec.')
    print('Begin detecting DE from %d circRNAs in %d vs %d samples.' % (circMat.shape[0],group1_num,group2_num))
    stime=time.time()
    decNum=stat(circMat,linearMat,group1_num,sampleNum,adjust_inc_len,adjust_skp_len,cutoff,groupPair,threads,options['-o'],options['--c2'],filterSample)
    etime=time.time()-stime
    print('Identify %d DEC with %f cutoff in %d seconds.' %(decNum,cutoff, etime))
##### functions part #############
def psi2PBSI(psi):
    pbsi=[]
    for i in psi:
        eachPbsi=[]
        eachPsi=i.split(',')
        for j in eachPsi:
            if j!='':
                eachPbsi.append(str(1-float(j)))
            else:
                eachPbsi.append('')
        pbsi.append(','.join(eachPbsi))
    return(pbsi)
def getMean(x):
    num=len(x)
    xSum=0
    for i in x:
        if i == 'nan':
            num=num-1
            continue
        try:
            tmp=float(i)
            xSum=xSum+tmp
        except:
            num=num-1
    if num ==0:
        return(0)
    else:
        return(xSum/num)
def get_delta(PBSI1,PBSI2):
    delta_pbsi=[]
    for i in range(len(PBSI1)):
        eachPbsi1_mean=getMean(PBSI1[i].split(','))
        eachPbsi2_mean=getMean(PBSI2[i].split(','))
        delta_pbsi.append(eachPbsi1_mean-eachPbsi2_mean)
    return(delta_pbsi)
def vec_remove_na_pair(inc1,skp1,inc2,skp2):
    res_inc1=[];res_skp1=[];res_inc2=[];res_skp2=[];
    for i in range(len(inc1)):
        if ((float(inc1[i])+float(skp1[i]))>0) & ((float(inc2[i])+float(skp2[i]))>0):
            res_inc1.append(inc1[i]);
            res_skp1.append(skp1[i]);
            res_inc2.append(inc2[i]);
            res_skp2.append(skp2[i]);
    return([res_inc1,res_skp1,res_inc2,res_skp2])
def vec_remove_na(inc,skp):
    res_inc=[];res_skp=[];
    for i in range(len(inc)):
        if (float(inc[i])+float(skp[i]))>0:
            res_inc.append(inc[i]);
            res_skp.append(skp[i]);
    return([res_inc,res_skp]);

def vec_remove_low_pair(inc1,skp1,inc2,skp2,num):
    res_inc1=[];res_skp1=[];res_inc2=[];res_skp2=[];
    for i in range(len(inc1)):
        if (float(skp1[i])>=num) & (float(skp2[i])>=num):
            res_inc1.append(inc1[i]);
            res_skp1.append(skp1[i]);
            res_inc2.append(inc2[i]);
            res_skp2.append(skp2[i]);
    return([res_inc1,res_skp1,res_inc2,res_skp2])
def vec_remove_low(inc,skp,num):
    res_inc=[];res_skp=[];
    for i in range(len(inc)):
        if float(skp[i])>=num:
            res_inc.append(inc[i]);
            res_skp.append(skp[i]);
    return([res_inc,res_skp]);

def checkInf(x):
    xt=[]
    for i in x:
        if i==float('inf') or i==float('-inf'):
            xt.append(float('nan'))
        else:
            xt.append(i)
    return(xt)
def stat(circMat,linearMat,group1_num,sampleNum,adjust_inc_len,adjust_skp_len,cutoff,groupPair,threads,output,isc2,filterSample):
    list_n_original_diff=[];psi_list_1=[];psi_list_2=[]
    for i in range(circMat.shape[0]):
        inc1=linearMat.iloc[i,3:(group1_num+3)].tolist()
        inc2=linearMat.iloc[i,(group1_num+3):(sampleNum+3)].tolist()
        skp1=circMat.iloc[i,3:(group1_num+3)].tolist()
        skp2=circMat.iloc[i,(group1_num+3):(sampleNum+3)].tolist()
        effective_inclusion_length=adjust_inc_len[i]
        effective_skipping_length=adjust_skp_len[i]
        inc1=vec2float(inc1);skp1=vec2float(skp1);inc2=vec2float(inc2);skp2=vec2float(skp2)
        inc1=checkInf(inc1);skp1=checkInf(skp1);inc2=checkInf(inc2);skp2=checkInf(skp2)
        if effective_inclusion_length <=0:
            effective_inclusion_length=readLength
        if groupPair:
            temp=vec_remove_na_pair(inc1,skp1,inc2,skp2);inc1_nona=temp[0];skp1_nona=temp[1];inc2_nona=temp[2];skp2_nona=temp[3];
            temp=vec_remove_low_pair(inc1_nona,skp1_nona,inc2_nona,skp2_nona,filterSample);inc1_nona=temp[0];skp1_nona=temp[1];inc2_nona=temp[2];skp2_nona=temp[3];
            if (len(inc1_nona)==0) | (len(inc2_nona)==0):
                psi_list_1.append('')
                psi_list_2.append('')
                list_n_original_diff.append([inc1,inc2,skp1,skp2,effective_inclusion_length,effective_skipping_length,0,i]);
            else:
                psi_list_1.append(adjustPsi(inc1,skp1,effective_inclusion_length,effective_skipping_length))
                psi_list_2.append(adjustPsi(inc2,skp2,effective_inclusion_length,effective_skipping_length))
                inc1=inc1_nona;skp1=skp1_nona;inc2=inc2_nona;skp2=skp2_nona;
                inc1=vecAddOne(inc1);skp1=vecAddOne(skp1);inc2=vecAddOne(inc2);skp2=vecAddOne(skp2);
                if (len(inc1)==1) |(len(inc2)==1):
                    list_n_original_diff.append([inc1,inc2,skp1,skp2,effective_inclusion_length,effective_skipping_length,0,i]);
                else:
                    list_n_original_diff.append([inc1,inc2,skp1,skp2,effective_inclusion_length,effective_skipping_length,1,i]);
        else:
            temp1=vec_remove_na(inc1,skp1);temp2=vec_remove_na(inc2,skp2);inc1_nona=temp1[0];skp1_nona=temp1[1];inc2_nona=temp2[0];skp2_nona=temp2[1];
            temp1=vec_remove_low(inc1_nona,skp1_nona,filterSample);temp2=vec_remove_low(inc2_nona,skp2_nona,filterSample);inc1_nona=temp1[0];skp1_nona=temp1[1];inc2_nona=temp2[0];skp2_nona=temp2[1];
            if (len(inc1_nona)==0) | (len(inc2_nona)==0) | (effective_inclusion_length==0):
                psi_list_1.append('')
                psi_list_2.append('')
                list_n_original_diff.append([inc1,inc2,skp1,skp2,effective_inclusion_length,effective_skipping_length,0,i])
            else:
                psi_list_1.append(adjustPsi(inc1,skp1,effective_inclusion_length,effective_skipping_length))
                psi_list_2.append(adjustPsi(inc2,skp2,effective_inclusion_length,effective_skipping_length))
                inc1=inc1_nona;skp1=skp1_nona;inc2=inc2_nona;skp2=skp2_nona;
                inc1=vecAddOne(inc1);skp1=vecAddOne(skp1);inc2=vecAddOne(inc2);skp2=vecAddOne(skp2)
                list_n_original_diff.append([inc1,inc2,skp1,skp2,effective_inclusion_length,effective_skipping_length,1,i])
    P=getrMATS_P(list_n_original_diff,cutoff,threads)
    FDR=myFDR(P)
    pbsi1=psi2PBSI(psi_list_1)
    pbsi2=psi2PBSI(psi_list_2)
    deltPBSI=get_delta(pbsi1,pbsi2)
    cjc1=circMat.iloc[:,3].map(str)
    cjc2=circMat.iloc[:,(group1_num+3)].map(str)
    ljc1=linearMat.iloc[:,3].map(str)
    ljc2=linearMat.iloc[:,(group1_num+3)].map(str)
    for i in range(4,group1_num+3):
        cjc1=cjc1+','+circMat.iloc[:,i].map(str)
    for i in range(group1_num+4,sampleNum+3):
        cjc2=cjc2+','+circMat.iloc[:,i].map(str)
    for i in range(4,group1_num+3):
        ljc1=ljc1+','+linearMat.iloc[:,i].map(str)
    for i in range(group1_num+4,sampleNum+3):
        ljc2=ljc2+','+linearMat.iloc[:,i].map(str)
    if isc2:
        matStat=pd.DataFrame({'chr':circMat.iloc[:,0],'start':circMat.iloc[:,1],'end':circMat.iloc[:,2],
        'cjc_1':cjc1.tolist(),'cjc_2':cjc2.tolist(),'cjc2_1':ljc1.tolist(),
        'cjc2_2':ljc2.tolist(),'adj_cjc_len':adjust_skp_len,'adj_cjc2_len':adjust_inc_len,
        'pbsi1':pbsi1,'pbsi2':pbsi2,'delta_pbsi':deltPBSI,'P':P,'FDR':FDR})
        matStat.to_csv(output,index=False,sep='\t',header=True)
    else:
        matStat=pd.DataFrame({'chr':circMat.iloc[:,0],'start':circMat.iloc[:,1],'end':circMat.iloc[:,2],
        'cjc_1':cjc1.tolist(),'cjc_2':cjc2.tolist(),'ljc_1':ljc1.tolist(),
        'ljc_2':ljc2.tolist(),'adj_cjc_len':adjust_skp_len,'adj_ljc_len':adjust_inc_len,
        'pbsi1':pbsi1,'pbsi2':pbsi2,'delta_pbsi':deltPBSI,'P':P,'FDR':FDR})
        matStat.to_csv(output,index=False,sep='\t',header=True)
    return(sum(np.array(FDR)<0.05))



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# import libraries
import subprocess
import re,os,warnings,math,scipy,itertools,logging,time,datetime,pysam,traceback
from scipy import stats
from multiprocessing import Pool
from scipy.optimize import fmin_cobyla
from scipy.optimize import fmin_l_bfgs_b
from math import log
import pandas as pd
import numpy as np
from numpy import *
np.seterr(divide='ignore',invalid='ignore')
rho=0.9
np.random.seed(1231);
########## functions ############
#binomial MLE optimization functions
def logit(x):
    if x<0.01:
        x=0.01
    if x>0.99:
        x=0.99
    return(log(x/(1-x)))

def myfunc_multivar(x,*args):
    psi1=args[0];psi2=args[1];var1=args[2];var2=args[3]
    #print('psi1');print(psi1);print('psi2');print(psi2)
    sum1=0;sum2=0
    for i in psi1:
        sum1+=pow(logit(i)-logit(x[0]),2)
    sum1=sum1/var1/2
    for i in psi2:
        sum2+=pow(logit(i)-logit(x[1]),2)
    sum2=sum2/var2/2
    return(sum1+sum2+0.1*0.5*(pow(rho,2))/(1-pow(rho,2))*(pow(stats.norm.ppf(x[0]),2)+pow(stats.norm.ppf(x[1]),2)-2*rho*stats.norm.ppf(x[0])*stats.norm.ppf(x[1])))

def myfunc_multivar_der(x,*args):
    psi1=args[0];psi2=args[1];var1=args[2];var2=args[3]
    sum1=0;sum2=0
    for i in psi1:
        sum1+=-2*(logit(i)-logit(x[0]))/x[0]/(1-x[0])
    sum1=sum1/var1/2
    res1=sum1+0.1*0.5*(pow(rho,2))/(1-pow(rho,2))*(2*stats.norm.ppf(x[0])-2*rho*stats.norm.ppf(x[1]))/stats.norm.pdf(stats.norm.ppf(x[0]))
    #print('1');print(x[1]);print(res1);print(stats.norm.pdf(stats.norm.ppf(x[0])))
    for i in psi2:
        sum2+=-2*(logit(i)-logit(x[1]))/x[1]/(1-x[1])
    sum2=sum2/var2/2
    res2=sum2+0.1*0.5*(pow(rho,2))/(1-pow(rho,2))*(2*stats.norm.ppf(x[1])-2*rho*stats.norm.ppf(x[0]))/stats.norm.pdf(stats.norm.ppf(x[1]))
    return(np.array([res1,res2]))

def myfunc_1(x, *args):
    psi1=args[0];psi2=args[1];var1=args[2];var2=args[3]
    sum1=0;sum2=0
    for i in psi1:
        sum1+=pow(logit(i)-logit(x+cutoff),2)
    sum1=sum1/var1/2
    for i in psi2:
        sum2+=pow(logit(i)-logit(x),2)
    sum2=sum2/var2/2
    return(sum1+sum2+0.1*0.5*(pow(rho,2))/(1-pow(rho,2))*(pow(stats.norm.ppf(x+cutoff),2)+pow(stats.norm.ppf(x),2)-2*rho*stats.norm.ppf(x+cutoff)*stats.norm.ppf(x)))

def myfunc_der_1(x, *args):
    psi1=args[0];psi2=args[1];var1=args[2];var2=args[3]
    sum1=0;sum2=0
    for i in psi1:
        sum1+=-2*(logit(i)-logit(x+cutoff))/(x+cutoff)/(1-x-cutoff)
    sum1=sum1/var1/2
    res1=sum1+0.1*0.5*(pow(rho,2))/(1-pow(rho,2))*(2*stats.norm.ppf(x+cutoff)-2*rho*stats.norm.ppf(x))/stats.norm.pdf(stats.norm.ppf(x+cutoff))
    for i in psi2:
        sum2+=-2*(logit(i)-logit(x))/x/(1-x)
    sum2=sum2/var2/2
    res2=sum2+0.1*0.5*(pow(rho,2))/(1-pow(rho,2))*(2*stats.norm.ppf(x)-2*rho*stats.norm.ppf(x+cutoff))/stats.norm.pdf(stats.norm.ppf(x))
    return(res1+res2)

def myfunc_2(x, *args):
    psi1=args[0];psi2=args[1];var1=args[2];var2=args[3]
    sum1=0;sum2=0
    for i in psi1:
        sum1+=pow(logit(i)-logit(x),2)
    sum1=sum1/var1/2
    for i in psi2:
        sum2+=pow(logit(i)-logit(x+cutoff),2)
    sum2=sum2/var2/2
    return(sum1+sum2+0.1*0.5*(pow(rho,2))/(1-pow(rho,2))*(pow(stats.norm.ppf(x+cutoff),2)+pow(stats.norm.ppf(x),2)-2*rho*stats.norm.ppf(x+cutoff)*stats.norm.ppf(x)))

def myfunc_der_2(x, *args):
    psi1=args[0];psi2=args[1];var1=args[2];var2=args[3]
    sum1=0;sum2=0
    for i in psi1:
        sum1+=-2*(logit(i)-logit(x))/(x)/(1-x)
    sum1=sum1/var1/2
    res1=sum1+0.1*0.5*(pow(rho,2))/(1-pow(rho,2))*(2*stats.norm.ppf(x+cutoff)-2*rho*stats.norm.ppf(x))/stats.norm.pdf(stats.norm.ppf(x+cutoff))
    for i in psi2:
        sum2+=-2*(logit(i)-logit(x+cutoff))/(x+cutoff)/(1-x-cutoff)
    sum2=sum2/var2/2
    res2=sum2+0.1*0.5*(pow(rho,2))/(1-pow(rho,2))*(2*stats.norm.ppf(x)-2*rho*stats.norm.ppf(x+cutoff))/stats.norm.pdf(stats.norm.ppf(x))
    return(res1+res2)

def myfunc_marginal_2der(x, args):
    I=args[0];S=args[1];beta=args[2];var=args[3]
    inclusion_length=args[4]
    skipping_length=args[5]
    #print('test x');print(x);print(var)
    temp1=1/pow(x,2)/pow(1-x,2)*(((2*x-1)*(logit(beta)-logit(x))-1)/var-1);#temp1=(2*x-1)/pow(x,2)/pow(1-x,2)*((logit(beta)-logit(x)-1/(2*x-1))/var+1)
    temp2=I*skipping_length*((2*inclusion_length+skipping_length)*x+skipping_length*(1-x))/pow(x,2)/pow(inclusion_length*x+skipping_length*(1-x),2)
    temp3=S*inclusion_length*((inclusion_length+2*skipping_length)*(1-x)+inclusion_length*x)/pow(1-x,2)/pow(inclusion_length*x+skipping_length*(1-x),2)
    #print('test');print(beta);print(x);print(var);print(temp1);print(temp2);print(temp3)
    return(temp1-temp2-temp3)
    
def myfunc_marginal(x, *args):
    beta=x
    I=args[0];S=args[1];psi=args[2];var=args[3]
    inclusion_length=args[4];skipping_length=args[5]
    sum=0
    for i in range(len(psi)):
        new_psi=inclusion_length*psi[i]/(inclusion_length*psi[i]+skipping_length*(1-psi[i]))
        f1=I[i]*log(new_psi)+S[i]*log(1-new_psi)-pow(logit(psi[i])-logit(beta),2)/2/var-log(psi[i])-log(1-psi[i])
        f1_2der=abs(myfunc_marginal_2der(psi[i],[I[i],S[i],beta,var,inclusion_length,skipping_length]))
        sum+=(0.5*log(abs(f1_2der)+0.00001)-f1)
    return(sum)

def myfunc_marginal_der(x, *args):
    beta=x
    I=args[0];S=args[1];psi=args[2];var=args[3]
    inclusion_length=args[4];skipping_length=args[5]
    sum=0
    for i in range(len(psi)):
        new_psi=inclusion_length*psi[i]/(inclusion_length*psi[i]+skipping_length*(1-psi[i]))
        f1_3der=1*(2*psi[i]-1)/pow(psi[i],2)/pow(1-psi[i],2)/beta/(1-beta)/var
        f1_2der=myfunc_marginal_2der(psi[i],[I[i],S[i],beta,var,inclusion_length,skipping_length])
        #print('test2');print(f1_2der)
        f1_der=(logit(psi[i])-logit(beta))/beta/(1-beta)/var
        f1=I[i]*log(new_psi)+S[i]*log(1-new_psi)-pow(logit(psi[i])-logit(beta),2)/2/var+log(psi[i])+log(1-psi[i])
        sum+=(0.5*f1_3der/(f1_2der)-f1_der)
    return(sum)

def myfunc_marginal_1(x, *args):
    beta2=x;beta1=x+cutoff
    I1=args[0];S1=args[1];psi1=args[2];var1=args[3]
    I2=args[4];S2=args[5];psi2=args[6];var2=args[7]
    inclusion_length=args[8];skipping_length=args[9]
    return(myfunc_marginal(beta1,I1,S1,psi1,var1,inclusion_length,skipping_length)+myfunc_marginal(beta2,I2,S2,psi2,var2,inclusion_length,skipping_length))

def myfunc_marginal_1_der(x, *args):
    beta2=x;beta1=x+cutoff
    I1=args[0];S1=args[1];psi1=args[2];var1=args[3]
    I2=args[4];S2=args[5];psi2=args[6];var2=args[7]
    inclusion_length=args[8];skipping_length=args[9]
    return(myfunc_marginal_der(beta1,I1,S1,psi1,var1,inclusion_length,skipping_length)+myfunc_marginal_der(beta2,I2,S2,psi2,var2,inclusion_length,skipping_length))

def myfunc_marginal_2(x, *args):
    beta2=x+cutoff;beta1=x
    I1=args[0];S1=args[1];psi1=args[2];var1=args[3]
    I2=args[4];S2=args[5];psi2=args[6];var2=args[7]
    inclusion_length=args[8];skipping_length=args[9]
    return(myfunc_marginal(beta1,I1,S1,psi1,var1,inclusion_length,skipping_length)+myfunc_marginal(beta2,I2,S2,psi2,var2,inclusion_length,skipping_length))

def myfunc_marginal_2_der(x, *args):
    beta2=x+cutoff;beta1=x
    I1=args[0];S1=args[1];psi1=args[2];var1=args[3]
    I2=args[4];S2=args[5];psi2=args[6];var2=args[7]
    inclusion_length=args[8];skipping_length=args[9]
    return(myfunc_marginal_der(beta1,I1,S1,psi1,var1,inclusion_length,skipping_length)+myfunc_marginal_der(beta2,I2,S2,psi2,var2,inclusion_length,skipping_length))
    
def myfunc_individual(x,*args):
    I=args[0];S=args[1];beta=args[2];var=args[3]
    inclusion_length=args[4]
    skipping_length=args[5]
    new_psi=inclusion_length*x/(inclusion_length*x+skipping_length*(1-x))
    return(-1*(I*log(new_psi)+S*log(1-new_psi)-(logit(x)-logit(beta))*(logit(x)-logit(beta))/2/var-log(x)-log(1-x)))

def myfunc_individual_der(x,*args):
    I=args[0];S=args[1];beta=args[2];var=args[3]
    inclusion_length=args[4]
    skipping_length=args[5]
    new_psi=inclusion_length*x/(inclusion_length*x+skipping_length*(1-x))
    new_psi_der=inclusion_length*skipping_length/pow(inclusion_length*x+skipping_length*(1-x),2)
    return(-1*(I/new_psi*new_psi_der-S/(1-new_psi)*new_psi_der-(logit(x)-logit(beta))/var/x/(1-x)-1/x+1/(1-x) ))

def myfunc_likelihood(x, args):
    I=args[0];S=args[1];beta=args[2];var=args[3];sum=0;N=I+S
    #return(-1*(-log(sqrt((I+S)*x*(1-x)))-(I-(I+S)*x)*(I-(I+S)*x)/2/((I+S)*x*(1-x))-log(sqrt(var))-(x-beta)*(x-beta)/2/var))
    #print('debug');print(N);print(var);print(x);print(beta)
    if N==0:
        return(0)
    return(-0.5*((I-N*x)*(I-N*x)/(N*x)+(S-N*(1-x))*(S-N*(1-x))/(N*(1-x)))-log(sqrt(var))-(x-beta)*(x-beta)/2/var)

def MLE_iteration_constrain(i1,i2,s1,s2,effective_inclusion_length,effective_skipping_length):
    psi1=vec2psi(i1,s1,effective_inclusion_length,effective_skipping_length);psi2=vec2psi(i2,s2,effective_inclusion_length,effective_skipping_length)
    iter_cutoff=1;iter_maxrun=100;count=0;previous_sum=0
    beta_0=sum(psi1)/len(psi1)
    beta_1=sum(psi2)/len(psi2)
    var1=10*scipy.var(np.array(psi1)-beta_0)
    var2=10*scipy.var(np.array(psi2)-beta_1)
    if var1<=0.01:
        var1=0.01
    if var2<=0.01:
        var2=0.01
    #print('var1');print(var1);print('var2');print(var2)
    while((iter_cutoff>0.01)&(count<=iter_maxrun)):
        count+=1
        #iteration of beta
        beta_0=sum(psi1)/len(psi1)
        beta_1=sum(psi2)/len(psi2)
        #print('var1');print(var1);print('var2');print(var2)
        #if abs(sum(psi1)/len(psi1)-sum(psi2)/len(psi2))>cutoff:
        if (sum(psi1)/len(psi1))>(sum(psi2)/len(psi2)):#minize psi2 if this is the case
            xopt = fmin_l_bfgs_b(myfunc_1,[sum(psi2)/len(psi2)],myfunc_der_1,args=[psi1,psi2,var1,var2],bounds=[[0.001,0.999-cutoff]],iprint=-1)
            theta2 = max(min(float(xopt[0]),1-cutoff),0);theta1=theta2+cutoff
        else:#minize psi1 if this is the case
            xopt = fmin_l_bfgs_b(myfunc_2,[sum(psi1)/len(psi1)],myfunc_der_2,args=[psi1,psi2,var1,var2],bounds=[[0.001,0.999-cutoff]],iprint=-1)
            theta1 = max(min(float(xopt[0]),1-cutoff),0);theta2=theta1+cutoff
        #print('constrain_1xopt');print('theta');print(theta1);print(theta2);print(xopt)
        #else:
        #    theta1=sum(psi1)/len(psi1);theta2=sum(psi2)/len(psi2)
        beta_0=theta1;beta_1=theta2
        #iteration of psi
        new_psi1=[];new_psi2=[];current_sum=0;likelihood_sum=0
        #print('constrain_2xopt')
        for i in range(len(psi1)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi1[i]],myfunc_individual_der,args=[i1[i],s1[i],beta_0,var1,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1)
            new_psi1.append(float(xopt[0]));current_sum+=float(xopt[1]);#print(xopt)
            #likelihood_sum+=myfunc_marginal(new_psi1[i],[i1[i],s1[i],beta_0,var1,effective_inclusion_length,effective_skipping_length])
        for i in range(len(psi2)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi2[i]],myfunc_individual_der,args=[i2[i],s2[i],beta_1,var2,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1)
            new_psi2.append(float(xopt[0]));current_sum+=float(xopt[1]);#print(xopt)
            #likelihood_sum+=myfunc_marginal(new_psi2[i],[i2[i],s2[i],beta_1,var2,effective_inclusion_length,effective_skipping_length])
        #print('new_psi[0]');print(new_psi1[0]);print(new_psi2[0])
        psi1=new_psi1;psi2=new_psi2
        #print('count');print(count);print('previous_sum');print(previous_sum);print('current_sum');print(current_sum)
        if count>1:
            iter_cutoff=abs(previous_sum-current_sum)
        previous_sum=current_sum
    #print('constrain');print(theta1);print(theta2);print(psi1);print(psi2);print(current_sum);print(likelihood_sum)
    #print(xopt)
    return([current_sum,[psi1,psi2,beta_0,beta_1,var1,var2]])
    #return([likelihood_sum,[psi1,psi2,beta_0,beta_1,var1,var2]])

def MLE_iteration(i1,i2,s1,s2,effective_inclusion_length,effective_skipping_length):
    psi1=vec2psi(i1,s1,effective_inclusion_length,effective_skipping_length);psi2=vec2psi(i2,s2,effective_inclusion_length,effective_skipping_length)
    iter_cutoff=1;iter_maxrun=100;count=0;previous_sum=0
    beta_0=sum(psi1)/len(psi1)
    beta_1=sum(psi2)/len(psi2)
    var1=10*scipy.var(np.array(psi1)-beta_0)
    var2=10*scipy.var(np.array(psi2)-beta_1)
    if var1<=0.01:
        var1=0.01
    if var2<=0.01:
        var2=0.01
    #print('var1');print(var1);print('var2');print(var2)
    while((iter_cutoff>0.01)&(count<=iter_maxrun)):
        count+=1
        #iteration of beta
        beta_0=sum(psi1)/len(psi1)
        beta_1=sum(psi2)/len(psi2)
        xopt=fmin_l_bfgs_b(myfunc_multivar,[beta_0,beta_1],myfunc_multivar_der,args=[psi1,psi2,var1,var2],bounds=[[0.01,0.99],[0.01,0.99]],iprint=-1)
        beta_0=float(xopt[0][0])
        beta_1=float(xopt[0][1])
        #print('unconstrain_1xopt');print(xopt)
        #print('theta');print(beta_0);print(beta_1);print('theta_end')
        #iteration of psi
        new_psi1=[];new_psi2=[];current_sum=0;likelihood_sum=0
        for i in range(len(psi1)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi1[i]],myfunc_individual_der,args=[i1[i],s1[i],beta_0,var1,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1)
            new_psi1.append(float(xopt[0]));current_sum+=float(xopt[1]);#print(xopt)
            #likelihood_sum+=myfunc_marginal(new_psi1[i],[i1[i],s1[i],beta_0,var1,effective_inclusion_length,effective_skipping_length])
        for i in range(len(psi2)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi2[i]],myfunc_individual_der,args=[i2[i],s2[i],beta_1,var2,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1)
            new_psi2.append(float(xopt[0]));current_sum+=float(xopt[1]);#print(xopt)
            #likelihood_sum+=myfunc_marginal(new_psi2[i],[i2[i],s2[i],beta_1,var2,effective_inclusion_length,effective_skipping_length])
        #print('new_psi[0]');print(new_psi1[0]);print(new_psi2[0])
        psi1=new_psi1;psi2=new_psi2
        #print('count');print(count);('previous_sum');print(previous_sum);print('current_sum');print(current_sum)
        if count>1:
            iter_cutoff=abs(previous_sum-current_sum)
        previous_sum=current_sum
    #print('unconstrain');print(beta_0);print(beta_0+beta_1);print(psi1);print(psi2);print(current_sum);print(likelihood_sum)
    #print(xopt)
    if count>iter_maxrun:
        return([current_sum,[psi1,psi2,0,0,var1,var2]])
    return([current_sum,[psi1,psi2,beta_0,beta_1,var1,var2]])
    #return([likelihood_sum,[psi1,psi2,beta_0,beta_1,var1,var2]])

def MLE_marginal_iteration(i1,i2,s1,s2,effective_inclusion_length,effective_skipping_length):
    #initial value
    psi1=vec2psi(i1,s1,effective_inclusion_length,effective_skipping_length);psi2=vec2psi(i2,s2,effective_inclusion_length,effective_skipping_length)
    iter_cutoff=1;iter_maxrun=100;count=0;previous_sum=0
    beta_0=sum(psi1)/len(psi1)
    beta_1=sum(psi2)/len(psi2)
    var1=10*scipy.var(np.array(psi1)-beta_0)
    var2=10*scipy.var(np.array(psi2)-beta_1)
    if var1<=0.01:
        var1=0.01
    if var2<=0.01:
        var2=0.01
    #print('var1');print(var1);print('var2');print(var2)
    
    #MLE of the full likelihood
    while((iter_cutoff>0.01)&(count<=iter_maxrun)):
        count+=1
        #iteration of beta
        beta_0=sum(psi1)/len(psi1)
        beta_1=sum(psi2)/len(psi2)
        xopt=fmin_l_bfgs_b(myfunc_multivar,[beta_0,beta_1],myfunc_multivar_der,args=[psi1,psi2,var1,var2],bounds=[[0.01,0.99],[0.01,0.99]],iprint=-1)
        beta_0=float(xopt[0][0])
        beta_1=float(xopt[0][1])
        #print('unconstrain_MLE_xopt');print(xopt)
        #print('theta');print(beta_0);print(beta_1);print('theta_end')
        #iteration of psi
        new_psi1=[];new_psi2=[];current_sum=0;likelihood_sum=0
        for i in range(len(psi1)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi1[i]],myfunc_individual_der,args=[i1[i],s1[i],beta_0,var1,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1)
            new_psi1.append(float(xopt[0]));current_sum+=float(xopt[1])
            #likelihood_sum+=myfunc_marginal(new_psi1[i],[i1[i],s1[i],beta_0,var1,effective_inclusion_length,effective_skipping_length])
        for i in range(len(psi2)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi2[i]],myfunc_individual_der,args=[i2[i],s2[i],beta_1,var2,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1)
            new_psi2.append(float(xopt[0]));current_sum+=float(xopt[1])
            #likelihood_sum+=myfunc_marginal(new_psi2[i],[i2[i],s2[i],beta_1,var2,effective_inclusion_length,effective_skipping_length])
        #print('new_psi[0]');print(new_psi1[0]);print(new_psi2[0])
        psi1=new_psi1;psi2=new_psi2
        #print('count');print(count);print('previous_sum');print(previous_sum);print('current_sum');print(current_sum)
        if count>1:
            iter_cutoff=abs(previous_sum-current_sum)
        previous_sum=current_sum
    #print('unconstrain');print(beta_0);print(beta_0+beta_1);print(psi1);print(psi2);print(current_sum);print(likelihood_sum)
    #print('unconstrain_psi_MLE');print(psi1);print(psi2)
    #print('unconstrain_beta_MLE');print(beta_0);print(beta_1)
    if count>iter_maxrun:
        return([current_sum,[psi1,psi2,0,0,var1,var2]])
    
    #MLE of the marginal likelihood
    iter_cutoff=1;iter_maxrun=100;count=0;previous_sum=0
    while((iter_cutoff>0.01)&(count<=iter_maxrun)):
        count+=1
        #print('unconstrain_MLE_marginal_value_1_der');print(myfunc_marginal_der(beta_0,i1,s1,psi1,var1,effective_inclusion_length,effective_skipping_length))
        #print('unconstrain_MLE_marginal_value_2_der');print(myfunc_marginal_der(beta_1,i2,s2,psi2,var2,effective_inclusion_length,effective_skipping_length))
        xopt1=fmin_l_bfgs_b(myfunc_marginal,[beta_0],myfunc_marginal_der,args=[i1,s1,psi1,var1,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1)
        xopt2=fmin_l_bfgs_b(myfunc_marginal,[beta_1],myfunc_marginal_der,args=[i2,s2,psi2,var2,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1)
        beta_0=float(xopt1[0][0])
        beta_1=float(xopt2[0][0])
        #print('unconstrain_1xopt');print(xopt1)
        #print('unconstrain_2xopt');print(xopt2)
        new_psi1=[];new_psi2=[];current_sum=0;likelihood_sum=0
        for i in range(len(psi1)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi1[i]],myfunc_individual_der,args=[i1[i],s1[i],beta_0,var1,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1)
            new_psi1.append(float(xopt[0]));current_sum+=float(xopt[1])
        for i in range(len(psi2)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi2[i]],myfunc_individual_der,args=[i2[i],s2[i],beta_1,var2,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1)
            new_psi2.append(float(xopt[0]));current_sum+=float(xopt[1])
        psi1=new_psi1;psi2=new_psi2
        if count>1:
            iter_cutoff=abs(previous_sum-current_sum)
        previous_sum=current_sum
    if count>iter_maxrun:
        return([current_sum,[psi1,psi2,0,0,var1,var2]])
    return([current_sum,[psi1,psi2,beta_0,beta_1,var1,var2]])

def MLE_marginal_iteration_constrain(i1,i2,s1,s2,effective_inclusion_length,effective_skipping_length):
    #initial value
    psi1=vec2psi(i1,s1,effective_inclusion_length,effective_skipping_length);psi2=vec2psi(i2,s2,effective_inclusion_length,effective_skipping_length)
    iter_cutoff=1;iter_maxrun=100;count=0;previous_sum=0
    beta_0=sum(psi1)/len(psi1)
    beta_1=sum(psi2)/len(psi2)
    var1=10*scipy.var(np.array(psi1)-beta_0)
    var2=10*scipy.var(np.array(psi2)-beta_1)
    if var1<=0.01:
        var1=0.01
    if var2<=0.01:
        var2=0.01
    #print('var1');print(var1);print('var2');print(var2)
    
    #MLE of the full likelihood
    while((iter_cutoff>0.01)&(count<=iter_maxrun)):
        count+=1
        #iteration of beta
        beta_0=sum(psi1)/len(psi1)
        beta_1=sum(psi2)/len(psi2)
        #print('var1');print(var1);print('var2');print(var2)
        #if abs(sum(psi1)/len(psi1)-sum(psi2)/len(psi2))>cutoff:
        if (sum(psi1)/len(psi1))>(sum(psi2)/len(psi2)):#minize psi2 if this is the case
            xopt = fmin_l_bfgs_b(myfunc_1,[sum(psi2)/len(psi2)],myfunc_der_1,args=[psi1,psi2,var1,var2],bounds=[[0.001,0.999-cutoff]],iprint=-1)
            theta2 = max(min(float(xopt[0]),1-cutoff),0);theta1=theta2+cutoff
        else:#minize psi1 if this is the case
            xopt = fmin_l_bfgs_b(myfunc_2,[sum(psi1)/len(psi1)],myfunc_der_2,args=[psi1,psi2,var1,var2],bounds=[[0.001,0.999-cutoff]],iprint=-1)
            theta1 = max(min(float(xopt[0]),1-cutoff),0);theta2=theta1+cutoff
        #print('constrain_1xopt');print('theta');print(theta1);print(theta2);print(xopt)
        #else:
        #    theta1=sum(psi1)/len(psi1);theta2=sum(psi2)/len(psi2)
        beta_0=theta1;beta_1=theta2
        #iteration of psi
        new_psi1=[];new_psi2=[];current_sum=0;likelihood_sum=0
        #print('constrain_2xopt')
        for i in range(len(psi1)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi1[i]],myfunc_individual_der,args=[i1[i],s1[i],beta_0,var1,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1)
            new_psi1.append(float(xopt[0]));current_sum+=float(xopt[1]);#print(xopt)
            #likelihood_sum+=myfunc_marginal(new_psi1[i],[i1[i],s1[i],beta_0,var1,effective_inclusion_length,effective_skipping_length])
        for i in range(len(psi2)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi2[i]],myfunc_individual_der,args=[i2[i],s2[i],beta_1,var2,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1)
            new_psi2.append(float(xopt[0]));current_sum+=float(xopt[1]);#print(xopt)
            #likelihood_sum+=myfunc_marginal(new_psi2[i],[i2[i],s2[i],beta_1,var2,effective_inclusion_length,effective_skipping_length])
        #print('new_psi[0]');print(new_psi1[0]);print(new_psi2[0])
        psi1=new_psi1;psi2=new_psi2
        #print('count');print(count);print('previous_sum');print(previous_sum);print('current_sum');print(current_sum)
        if count>1:
            iter_cutoff=abs(previous_sum-current_sum)
        previous_sum=current_sum
    if count>iter_maxrun:
        return([current_sum,[psi1,psi2,0,0,var1,var2]])
    #print('constrain');print(theta1);print(theta2);print(psi1);print(psi2);print(current_sum);print(likelihood_sum)
    #print(xopt)
    
    #MLE of the marginal likelihood
    iter_cutoff=1;iter_maxrun=100;count=0;previous_sum=0
    while((iter_cutoff>0.01)&(count<=iter_maxrun)):
        count+=1
        if (sum(psi1)/len(psi1))>(sum(psi2)/len(psi2)):#minize psi2 if this is the case
            xopt = fmin_l_bfgs_b(myfunc_marginal_1,[beta_1],myfunc_marginal_1_der,args=[i1,s1,psi1,var1,i2,s2,psi2,var2,effective_inclusion_length,effective_skipping_length],bounds=[[0.001,0.999-cutoff]],iprint=-1)
            beta_1 = max(min(float(xopt[0]),1-cutoff),0);beta_0=beta_1+cutoff
        else:#minize psi1 if this is the case
            xopt = fmin_l_bfgs_b(myfunc_marginal_2,[beta_0],myfunc_marginal_2_der,args=[i1,s1,psi1,var1,i2,s2,psi2,var2,effective_inclusion_length,effective_skipping_length],bounds=[[0.001,0.999-cutoff]],iprint=-1)
            beta_0 = max(min(float(xopt[0]),1-cutoff),0);beta_1=beta_0+cutoff
        #print('constrain_xopt');print(xopt)
        new_psi1=[];new_psi2=[];current_sum=0;likelihood_sum=0
        for i in range(len(psi1)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi1[i]],myfunc_individual_der,args=[i1[i],s1[i],beta_0,var1,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1)
            new_psi1.append(float(xopt[0]));current_sum+=float(xopt[1]);#print(xopt)
        for i in range(len(psi2)):
            xopt = fmin_l_bfgs_b(myfunc_individual,[psi2[i]],myfunc_individual_der,args=[i2[i],s2[i],beta_1,var2,effective_inclusion_length,effective_skipping_length],bounds=[[0.01,0.99]],iprint=-1)
            new_psi2.append(float(xopt[0]));current_sum+=float(xopt[1]);#print(xopt)
        psi1=new_psi1;psi2=new_psi2
        if count>1:
            iter_cutoff=abs(previous_sum-current_sum)
        previous_sum=current_sum
    if count>iter_maxrun:
        return([current_sum,[psi1,psi2,0,0,var1,var2]])

    return([current_sum,[psi1,psi2,beta_0,beta_1,var1,var2]])
    #return([likelihood_sum,[psi1,psi2,beta_0,beta_1,var1,var2]])

    
#Random Sampling Function
def likelihood_test(i1,i2,s1,s2,effective_inclusion_length,effective_skipping_length,flag,id):
    #print('testing'+str(id))
    if flag==0:
        return([1,1])
    else:
        res=MLE_marginal_iteration(i1,i2,s1,s2,effective_inclusion_length,effective_skipping_length)
        if abs(res[1][2]-res[1][3])<=cutoff:
            #print('1<=cutoff');print(res);print((res[1][2]-res[1][3]))
            return([1,res[0]])
        else:
            res_constrain=MLE_marginal_iteration_constrain(i1,i2,s1,s2,effective_inclusion_length,effective_skipping_length)
            #print('2>cutoff');print(res);print(res_constrain)
            #print(abs(res_constrain[0]-res[0]));print('2end')
            #print(abs(res_constrain[2]-res[2]));print('2end_marginal');            
            if len(i1)<=3:
                return([1-scipy.stats.chi2.cdf(2*(abs(res_constrain[0]-res[0])),1)])
            else:
                return([1-scipy.stats.chi2.cdf(2*(abs(res_constrain[0]-res[0])),1)])

#MultiProcessorFunction
def MultiProcessorPool(n_original_diff):
    i1=n_original_diff[0];i2=n_original_diff[1];s1=n_original_diff[2];s2=n_original_diff[3]
    effective_inclusion_length=n_original_diff[4];effective_skipping_length=n_original_diff[5]
    flag=n_original_diff[6];id=n_original_diff[7]
    P=likelihood_test(i1,i2,s1,s2,effective_inclusion_length,effective_skipping_length,flag,id)
    return(P)

#Function for vector handling
def vec2float(vec):
    res=[]
    for i in vec:
        res.append(float(i))
    return(res)

#add 1 in both inclusion and skipping counts for robustness in small sample size
def vecAddOne(vec):
    res=[]
    for i in vec:
        res.append(i+1)
    return(res)

def vecprod(vec):
    res=1
    for i in vec:
        res=res*i
    return(res)

def vecadd(vec1,vec2):
    res=[]
    for i in range(len(vec1)):
        res.append(vec1[i]+vec2[i])
    return(res)
    
def vec2remove0psi(inc,skp):
    res1=[];res2=[]
    for i in range(len(inc)):
        if (inc[i]!=0) | (skp[i]!=0):
            res1.append(inc[i]);res2.append(skp[i])
    return([res1,res2])

def vec2psi(inc,skp,effective_inclusion_length,effective_skipping_length):
    psi=[]
    inclusion_length=effective_inclusion_length
    skipping_length=effective_skipping_length
    for i in range(len(inc)):
        if (float(inc[i])+float(skp[i]))==0:
            psi.append(0.5)
        else:
            psi.append(float(inc[i])/inclusion_length/(float(inc[i])/inclusion_length+float(skp[i])/skipping_length))
    return(psi)

def vec210(vec):
    res=[]
    for i in vec:
        if i>0:
            res.append(1)
        else:
            res.append(-1)
    return(res)

def myttest(vec1,vec2):
    if (len(vec1)==1) & (len(vec2)==1):
        res=stats.ttest_ind([vec1[0],vec1[0]],[vec2[0],vec2[0]])
    else:
        res=stats.ttest_ind(vec1,vec2)
    return(res)
def myorder(p,reverse):
    p_dict={}
    #print('len(p)');print(len(p))
    for index in range(len(p)):
        if not(p[index] in p_dict):
            p_dict[p[index]]=[]
        p_dict[p[index]].append(index+1)
    #print('p_dict');print(p_dict)
    o_index=sorted(p_dict,reverse=reverse);o=[]
    #print('o_index');print(o_index)
    for index in o_index:
        for this_p in p_dict[index]:
            o.append(this_p)
    return(o)
def mycummin(p):
        res=[]
        for i in range(len(p)):
            res.append(min(p[:(i+1)]))
        return(res)

def myFDR(p):
    lp=len(p)
    i=range(lp,0,-1)
    o=myorder(p,True)
    ro=myorder(o,False)
    p_new=[]
    for index in range(len(o)):
        p_new.append(p[o[index]-1]*lp/i[index])
    p_mycummin=mycummin(p_new)
    p_mycummin_new=[]
    for index in range(len(p_mycummin)):
        p_mycummin_new.append(min(p_mycummin[index],1))
    res=[]
    for index in range(len(p_mycummin_new)):
        res.append(p_mycummin_new[ro[index]-1])
    return(res)

def adjustPsi(inc,skp,effective_inclusion_length,effective_skipping_length):
    #if effective_inclusion_length == 0:
        #print ([inc,skp,effective_inclusion_length,effective_skipping_length])
    inc_adj=np.array(inc)/effective_inclusion_length
    skp_adj=np.array(skp)/effective_skipping_length
    return(','.join([str(i) for i in (inc_adj/(inc_adj+skp_adj)).tolist()]))

def myand(l1,l2):
    return([i for i in range(len(l1)) if l1[i] & l2[i] ==True])

    ######## end of functions ##############
def getrMATS_P(list_n_original_diff,mycutoff,threads):
    global cutoff
    cutoff=mycutoff
    probability=[]
    if threads>1:
        pool=Pool(processes=threads)
        probability=pool.map(MultiProcessorPool,list_n_original_diff)
        pool.close()
        pool.join()
    else:
        for i in range(len(list_n_original_diff)):
            probability.append(MultiProcessorPool(list_n_original_diff[i]))
    P=pd.DataFrame(probability).iloc[:,0].tolist()
    return(P)
